import argparse
import os
import sys

from PyQt5.QtCore import Qt, QObject, pyqtSignal
from PyQt5.QtGui import QPen
from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QMainWindow, QHBoxLayout, QWidget, QTreeWidgetItem, QSplitter, QGraphicsRectItem, QTabWidget, QLabel, QVBoxLayout, \
    QSizePolicy, QFormLayout, QCheckBox, QGroupBox
from easyconfig2.easyconfig import EasyConfig2
from swikv4.widgets.swik_basic_widget import SwikBasicWidget

from qrgrader.utils import makedir
from qrgrader.code import Code
from qrgrader.code_set import CodeSet
from qrgrader.common import check_workspace, get_workspace_paths, Questions, Nia, StudentsData, get_prefix
from qrgrader.pdf_tree import PDFTree
from qrgrader.rubric import Rubric


class Mark(QGraphicsRectItem):
    class Signal(QObject):
        double_click = pyqtSignal(object)

    def __init__(self, code):
        super().__init__()
        self.signal = Mark.Signal()
        self.code = code
        self.setRect(code.x, code.y, code.w, code.h)
        if code.marked:
            self.setPen(QPen(Qt.red, 2))
        else:
            self.setPen(QPen(Qt.transparent, 2))

    def mouseDoubleClickEvent(self, event):
        self.signal.double_click.emit(self.code)


class MainWindow(QMainWindow):
    def __init__(self, schema_filenames):
        super().__init__()

        makedir(os.path.expanduser("~") + os.sep + ".config/qrgrader/")
        self.config = EasyConfig2(filename=os.path.expanduser("~") + os.sep + ".config/qrgrader/qrgrader.yaml")
        self.cfg_geometry = self.config.root().addPrivate("geometry", default=[0, 0, 1200, 1000, False])
        self.config.load()

        self.current_exam = None
        self.detected = CodeSet()
        self.changed = CodeSet()

        # Rubrics
        self.rubrics = []
        self.rubrics_files = schema_filenames
        self.rubrics_labels = []
        self.rubrics_cb = []

        (self.dir_workspace,
         self.dir_data,
         self.dir_scanned, _,
         self.dir_xls,
         self.dir_publish, _) = get_workspace_paths(os.getcwd())

        self.prefix = get_prefix()
        self.xls_questions = Questions(self.dir_xls + self.prefix + "questions.csv")
        self.xls_nia = Nia(self.dir_xls + self.prefix + "nia.csv")
        self.xls_data = StudentsData(self.dir_xls + self.prefix + "data.csv")

        self.central_widget = QWidget()
        self.main_layout = QVBoxLayout()
        self.central_widget.setLayout(self.main_layout)
        self.setCentralWidget(self.central_widget)
        self.pdf_tree = PDFTree()
        self.pdf_tree.setColumnCount(4)
        self.swik = SwikBasicWidget()
        self.swik.view.document_ready.connect(self.load_finished)
        self.splitter = QSplitter()

        # Prepare Details layout
        self.name_lbl = QLabel()
        self.name_lbl.setMinimumWidth(300)
        self.name_lbl.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.nia_lbl = QLabel()
        self.nia_lbl.setMinimumWidth(100)
        self.nia_lbl.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.group_lbl = QLabel()
        self.group_lbl.setMinimumWidth(100)
        self.group_lbl.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)

        details_layout = QHBoxLayout()
        details_layout.setAlignment(Qt.AlignLeft)
        details_layout.addWidget(QLabel("Name:"))
        details_layout.addWidget(self.name_lbl)
        details_layout.addWidget(QLabel("NIA:"))
        details_layout.addWidget(self.nia_lbl)
        details_layout.addWidget(QLabel("Group:"))
        details_layout.addWidget(self.group_lbl)

        self.main_layout.addWidget(self.splitter)
        self.main_layout.addLayout(details_layout)

        self.rubrics_tabs = QTabWidget()

        helper = QWidget()
        helper.setLayout(QVBoxLayout())

        self.scores_layout = QFormLayout()
        framebox = QGroupBox("Scores")
        framebox.setLayout(self.scores_layout)

        helper.layout().addWidget(framebox)
        helper.layout().addWidget(self.rubrics_tabs)

        self.quiz_score_lbl = QLabel("0")
        self.total_score_lbl = QLabel("0")
        self.quiz_cb = QCheckBox("Quiz")
        self.quiz_cb.setChecked(True)
        self.quiz_cb.stateChanged.connect(self.score_checkbox_changed)
        self.scores_layout.addRow(self.quiz_cb, self.quiz_score_lbl)
        self.scores_layout.addRow("Total: ", self.total_score_lbl)

        self.splitter.addWidget(self.pdf_tree)
        self.splitter.addWidget(self.swik)
        self.splitter.addWidget(helper)

        self.setWindowTitle("QR Grader")

        self.load_tables()

        self.load_detected()
        self.load_schemas()

        self.populate_pdf_tree()
        self.pdf_tree.currentItemChanged.connect(self.pdf_tree_selection_changed)
        if self.pdf_tree.topLevelItemCount() > 0:
            self.pdf_tree.setCurrentItem(self.pdf_tree.topLevelItem(0))

        x, y, w, h, fullscreen = self.cfg_geometry.get()
        if fullscreen:
            self.showFullScreen()
        else:
            self.setGeometry(x, y, w, h)

        self.show()

    def closeEvent(self, a0):
        self.cfg_geometry.set(
            [self.geometry().x(), self.geometry().y(), self.geometry().width(), self.geometry().height(),
             self.isFullScreen()])
        self.config.save()

    def load_detected(self):
        self.detected.load(self.dir_data + self.prefix + "detected.csv")

    def load_tables(self):

        if not self.xls_questions.load():
            print("ERROR: questions.csv file nos present")
            sys.exit(1)

        if not self.xls_nia.load():
            print("ERROR: nia.csv file nos present")
            sys.exit(1)

        if not self.xls_data.load():
            print("WARNING: data.csv file nos present")

    def load_schemas(self):
        for filename in self.rubrics_files:
            name = os.path.basename(filename).replace(".scm", "")
            r1 = Rubric(filename, self.dir_xls)
            r1.score_changed.connect(self.rubric_score_changed)
            r1.goto_next.connect(self.goto_next)
            self.rubrics_tabs.addTab(r1, name)
            self.rubrics.append(r1)

            label = QLabel("0")
            rubric_cb = QCheckBox(name + ":")
            rubric_cb.setChecked(True)
            rubric_cb.stateChanged.connect(self.score_checkbox_changed)

            self.rubrics_labels.append(label)
            self.rubrics_cb.append(rubric_cb)

            self.scores_layout.insertRow(1, rubric_cb, label)

    def goto_next(self):
        current = self.pdf_tree.currentItem()
        if current is not None:
            index = self.pdf_tree.indexOfTopLevelItem(current)
            if index < self.pdf_tree.topLevelItemCount() - 1:
                self.pdf_tree.setCurrentItem(self.pdf_tree.topLevelItem(index + 1))

    def score_checkbox_changed(self, state):
        self.update_all_pdf_tree_scores()
        self.update_scores_layout()

    def rubric_score_changed(self, rubric, exam_id):
        self.update_scores_layout()
        self.update_pdf_tree_score()

    def update_scores_layout(self):
        total = 0
        quiz_score = self.get_quiz_score(self.current_exam)
        self.quiz_score_lbl.setText(str(quiz_score))
        if self.quiz_cb.isChecked():
            total += quiz_score

        for index, r in enumerate(self.rubrics):
            value = r.compute_score(self.current_exam)
            self.rubrics_labels[index].setText(str(value))
            if self.rubrics_cb[index].isChecked():
                total += value

        self.total_score_lbl.setText(str(total))
        return total

    def pdf_tree_selection_changed(self, current, previous):

        self.pdf_tree.set_enabled(False)
        if previous is not None:
            for rubric in self.rubrics:
                rubric.push(int(previous.text(1)))

        self.current_exam = int(current.text(1))

        ratio = self.swik.view.get_ratio()
        rubric = self.rubrics_tabs.currentWidget()
        index = rubric.get_page() if rubric is not None else 0

        self.swik.open(f"{self.dir_publish}{self.current_exam}.pdf", ratio=ratio, index=index)

    def load_finished(self):

        self.process_exam()

        nia = self.xls_nia.get_nia(self.current_exam)
        self.nia_lbl.setText(str(nia))
        self.name_lbl.setText(str(self.xls_data.get_name(nia)))
        self.group_lbl.setText(str(self.xls_data.get_group(nia)))

        for rubric in self.rubrics:
            rubric.pull(self.current_exam)

        self.update_scores_layout()

        for index in range(self.pdf_tree.topLevelItemCount()):
            item = self.pdf_tree.topLevelItem(index)
            exam_id = int(item.text(1))
            if len(self.get_multiple_marks(exam_id)) > 0:
                item.setText(2, "!")
            else:
                item.setText(2, "")

        self.pdf_tree.set_enabled(True)

    def populate_pdf_tree(self):
        files = os.listdir(self.dir_publish)
        files = sorted([f.replace(".pdf", "") for f in files if f.endswith(".pdf") and f.replace(".pdf", "").isdigit()])

        for f in files:
            score = self.get_quiz_score(int(f))
            item = QTreeWidgetItem(["", f, str(score)])
            self.pdf_tree.addTopLevelItem(item)

        self.update_all_pdf_tree_scores()

    def process_exam(self):
        marks = [x for x in self.swik.view.items() if isinstance(x, Mark)]

        while len(marks) > 0:
            mark = marks.pop()
            self.swik.view.scene().removeItem(mark)

        marks = {}
        for index in range(self.swik.renderer.get_document_length()):
            exam_id = self.current_exam % 1000
            page_codes = self.detected.select(exam=exam_id, pdf_page=index + 1)
            type_a = page_codes.select(type=Code.TYPE_A)
            for code in type_a:
                r = Mark(code)
                r.signal.double_click.connect(self.code_clicked)
                r.setParentItem(self.swik.view.get_page(index))
                marks[code] = r
                if code.marked:
                    if self.xls_questions.get_value(code.question, code.answer) > 0:
                        r.setPen(QPen(Qt.green, 2))
                    else:
                        r.setPen(QPen(Qt.red, 2))

            type_n = page_codes.select(type=Code.TYPE_N)
            for code in type_n:
                r = Mark(code)
                r.signal.double_click.connect(self.code_clicked)
                r.setParentItem(self.swik.view.get_page(index))
                if code.marked:
                    r.setPen(QPen(Qt.blue, 2))

        yellow = self.get_multiple_marks(exam_id)
        for answer in yellow:
            marks[answer].setPen(QPen(Qt.yellow, 2))

    def get_full_score(self, exam_id):
        total = 0
        for index, r in enumerate(self.rubrics):
            if self.rubrics_cb[index].isChecked():
                total += r.compute_score(exam_id)
        if self.quiz_cb.isChecked():
            total += self.get_quiz_score(exam_id)

        return total

    def update_all_pdf_tree_scores(self):
        for index in range(self.pdf_tree.topLevelItemCount()):
            item = self.pdf_tree.topLevelItem(index)
            score = self.get_full_score(int(item.text(1)))
            item.setText(3, str(score))

    def update_pdf_tree_score(self):
        for index in range(self.pdf_tree.topLevelItemCount()):
            item = self.pdf_tree.topLevelItem(index)
            if int(item.text(1)) == self.current_exam:
                score = self.get_full_score(self.current_exam)
                item.setText(3, str(score))
                break

    def get_quiz_score(self, exam_id):
        score = 0
        exam_codes = self.detected.select(exam=exam_id % 1000, type=Code.TYPE_A)
        for code in exam_codes:
            if code.marked:
                score += self.xls_questions.get_value(code.question, code.answer)
        return score

    def get_multiple_marks(self, exam_id):
        yellow = []
        exam_id = exam_id % 1000
        type_a = self.detected.select(exam=exam_id, type=Code.TYPE_A)
        for question in type_a.get_questions():
            answers = type_a.select(question=question)
            marked = sum([1 for x in answers if x.marked])
            if marked > 1:
                for answer in answers:
                    if answer.marked:
                        yellow.append(answer)
        return yellow

    def code_clicked(self, code):
        self.changed.append(code)
        self.changed.save(self.dir_data + self.prefix + "changed.csv")
        code.marked = not code.marked
        self.process_exam()
        self.update_scores_layout()
        self.update_pdf_tree_score()
        self.detected.save(self.dir_data + self.prefix + "detected.csv")


def main():
    app = QApplication(sys.argv)

    if not check_workspace():
        print("ERROR: qrgrader must be run from the workspace root")
        sys.exit(1)

    parser = argparse.ArgumentParser(description='qrgui.py')
    parser.add_argument('-s', '--schema', help='Schema to be used', default=[], action="append")
    parser.add_argument('-c', '--create', help="Create schema if doesn't exist", action="store_true")
    args = vars(parser.parse_args())

    filenames = []
    for schema in args["schema"]:
        if schema.endswith(".yaml"):
            print("WARNING: schema MUST NOT be a yaml file.")
            sys.exit(1)

        filename = schema.replace(".scm", "") + ".scm"

        if not os.path.exists(filename):
            if args["create"]:
                print("Creating schema", filename)
                with open(filename, "w", encoding='utf-8') as f:
                    f.write("{}\n")
            else:
                print(f"ERROR: schema {filename} not found")
                sys.exit(1)
        filenames.append(filename)

    main = MainWindow(filenames)
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import csv
import os

import yaml
from PyQt5.QtCore import Qt, pyqtSignal, QTimer
from PyQt5.QtGui import QDrag, QPixmap, QPainter
from PyQt5.QtWidgets import (QListWidget,
                             QAbstractItemView, QListWidgetItem, QMenu, QMessageBox,
                             QInputDialog, QColorDialog)

from qrgrader.dialogs import ButtonEditDialog, RubricEditDialog
from qrgrader.buttons import StepButton, Shortcut, Button, TextButton, StateButton, Separator, CutterButton, MultiplierButton


class Rubric(QListWidget):
    score_changed = pyqtSignal(object, int)
    goto_next = pyqtSignal()

    def __init__(self, schema_filename, dir_xls):
        super().__init__()

        self.exam_id = None
        self.config = {}
        self.scores = {}
        self.schema_dictionary = {}
        self.schema_filename = schema_filename
        name = self.schema_filename.split(".")[0]
        self.scores_filename = name + ".yaml"
        self.xls_filename = dir_xls + name + ".csv"
        self.current_exam_id = None
        self.modified = False

        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.setMinimumWidth(115)

        # self.schemaChanged.connect(self.save_schema) # type: ignore
        self.customContextMenuRequested.connect(self.button_list_right_click)
        self.populate()
        self.load_scores()

    def load_scores(self):
        if os.path.exists(self.scores_filename):
            with open(self.scores_filename) as file:
                self.scores = yaml.safe_load(file)

    def push(self, exam_id):
        if exam_id is not None:
            self.store(exam_id)
            self.save_scores()

    def pull(self, exam_id):
        self.exam_id = exam_id
        self.retrieve(self.exam_id)

    # def exam_changed(self, exam_id, prev_exam_id):
    #     self.exam_id = exam_id
    #     if prev_exam_id is not None:
    #         self.store(prev_exam_id)
    #     self.save_scores()
    #
    #     self.retrieve(self.exam_id)

    def get_page(self):
        return self.config.get("page", 1) - 1

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_Down:
            self.goto_next.emit()
        else:
            super().keyPressEvent(e)

    def populate(self):
        if os.path.exists(self.schema_filename):
            # Load schema
            with open(self.schema_filename, newline='') as csvfile:
                content = yaml.full_load(csvfile)

                self.config = content.get("config", {})
                self.schema_dictionary = content.get("buttons", {})

                # Create lateral panel and buttons
                for button_name in self.schema_dictionary:
                    button_config = self.schema_dictionary[button_name]

                    if button_config.get("type") == 'button':
                        button = StepButton(button_name, **button_config)
                        button.score_changed.connect(self.button_clicked)  # type: ignore
                    elif button_config.get("type") == 'text':
                        button = TextButton(button_name, **button_config)
                    elif button_config.get("type") == 'cutter':
                        button = CutterButton(button_name, **button_config)
                        button.score_changed.connect(self.button_clicked)
                    elif button_config.get("type") == 'multiplier':
                        button = MultiplierButton(button_name, **button_config)
                        button.score_changed.connect(self.button_clicked)
                    elif button_config.get("type") == 'shortcut':
                        button = Shortcut(button_name, **button_config)
                        button.clicked.connect(self.shortcut_activated)
                    elif button_config.get("type") == 'separator':
                        button = Separator(button_name, **button_config)
                    else:
                        button = None

                    if button is not None:
                        # Create Item in ListWidget
                        item = QListWidgetItem()
                        self.addItem(item)
                        self.setItemWidget(item, button)
                        item.setSizeHint(button.sizeHint())

    def save_schema(self):
        buttons = {}
        for b in self.filter_buttons(Button):
            buttons[b.get_name()] = b.get_config()

        with open(self.schema_filename, "w", encoding='utf-8') as f:
            schema = {"buttons": buttons, "config": self.config}
            yaml.dump(schema, f, sort_keys=False)

    def button_clicked(self):
        self.store(self.exam_id)
        self.score_changed.emit(self, self.exam_id)
        button = self.sender()
        if button.get_click_next() and button.is_checked():
            QTimer.singleShot(500, self.goto_next.emit)

    def add_button(self):

        button = self.get_dialog()

        if button is None:
            return

        # Create Item in ListWidget
        item = QListWidgetItem()
        self.addItem(item)

        self.setItemWidget(item, button)
        item.setSizeHint(button.sizeHint())
        self.save_schema()

    def get_dialog(self, button=None):

        dialog = ButtonEditDialog(self, button)

        if not dialog.exec():
            return None

        name, kind, config = dialog.get()

        self.schema_dictionary[name] = config

        if kind == 'button':
            button = StepButton(name, **config)
            button.score_changed.connect(self.button_clicked)
        elif kind == 'text':
            button = TextButton(name, **config)
        elif kind == 'cutter':
            button = CutterButton(name, **config)
            button.score_changed.connect(self.button_clicked)
        elif kind == 'multiplier':
            button = MultiplierButton(name, **config)
            button.score_changed.connect(self.button_clicked)
        elif kind == 'separator':
            button = Separator(name, **config)
        else:
            button = None

        return button

    def edit_button(self, position):
        item = self.item(position)
        widget = self.itemWidget(item)
        prev_name = widget.get_name()

        button = self.get_dialog(widget)

        if button is None:
            return

        self.setItemWidget(item, button)

        # Rename the key in the dictionary
        if prev_name is not None:
            for exam_id in self.scores:
                if prev_name in self.scores[exam_id].keys():
                    self.scores[exam_id][button.get_name()] = self.scores[exam_id].pop(prev_name)

        self.save_schema()

    def button_list_right_click(self, pos):
        menu = QMenu()
        item = self.item(self.currentRow())
        widget = self.itemWidget(item)

        if isinstance(widget, StateButton):
            menu.addAction("Edit", lambda: self.edit_button(self.currentRow()))
            menu.addAction("Duplicate", lambda: self.duplicate_button(self.currentRow()))
            menu.addAction("Remove", lambda: self.delete_button(self.currentRow()))
            menu.addAction("Add Comment", lambda: self.add_comment(self.currentRow()))
            menu.addSeparator()
        elif isinstance(widget, Shortcut):
            menu.addAction("Remove", lambda: self.remove_shortcut(self.currentRow()))
            menu.addAction("Change Color", lambda: self.set_shortcut_color(self.currentRow()))
            menu.addSeparator()
        menu.addAction("Edit Rubric Config", self.edit_rubric_config)
        menu.addAction("Add Button", self.add_button)
        menu.addAction("Add Shortcut", self.add_shortcut)

        menu.exec(self.mapToGlobal(pos))
        self.clearSelection()
        self.clearFocus()

    def compute_score(self, exam_id):
        this_exam_data = self.scores.get(int(exam_id), {})

        total = 0
        points = 0
        for button in self.filter_buttons(StepButton):
            this_button_data = this_exam_data.get(button.get_name(), {})
            value = this_button_data.get("value", 0)
            if value >= 0:
                value = value * button.get_full_value() / 100.0
                points = points + value
            total = total + button.get_full_value()

        cut = 1
        for button in self.filter_buttons(CutterButton):
            cut = min(cut, button.get_percent())

        points = min(points, total * cut)

        multiplier = 1
        for button in self.filter_buttons(MultiplierButton):
            multiplier = button.get_percent()

        points = points * multiplier

        return points

    def save_scores(self):
        with open(self.scores_filename, "w", encoding='utf-8') as file:
            yaml.dump(self.scores, file)

        self.save_xls()

    def save_xls(self):

        with open(self.xls_filename, "w", encoding='utf-8') as f:
            row = "EXAM_ID"
            for button_name, _ in self.schema_dictionary.items():
                row += "\t" + button_name

            f.write(row + "\n")

            row = " "
            for button_name, button_config in self.schema_dictionary.items():
                if button_config.get("type") == 'button':
                    row += "\t" + "{:.2f}".format(button_config.get("full_value", 1))
                elif button_config.get("type") == 'text':
                    row += "\t "
                else:
                    row += "\t?"

            f.write(row + "\n")

            for exam_id, exam_score_items in self.scores.items():
                row = str(exam_id)
                for button_name, button_config in self.schema_dictionary.items():

                    button_type = button_config.get("type")
                    button_state = exam_score_items.get(button_name)  # value and comment or text

                    if button_state is not None:
                        if button_type == 'button':
                            value = button_state.get("value")
                            value = " " if value == -1 else round(value / 100, 2)
                        elif button_type == 'text':
                            value = button_state.get("text")
                        else:
                            value = "?"
                    else:
                        value = " "

                    row += "\t {:4s}".format(str(value))

                f.write(row + "\n")

    def shortcut_activated(self):
        buttons = self.sender().get_buttons()

        for b in self.filter_buttons(StepButton):  # type: StepButton
            b.button.setChecked(b.get_name() in buttons)
            b.clicked()

    def add_shortcut(self):
        text, ok = QInputDialog.getText(self, "Shortcut", "Name:")
        if ok:

            buttons = []
            for b in self.filter_buttons(StepButton):  # type: StepButton
                if b.button.isChecked():
                    buttons.append(b.get_name())

            b2 = Shortcut(text, buttons=buttons)
            b2.clicked.connect(self.shortcut_activated)  # type: ignore

            item = QListWidgetItem()
            self.addItem(item)
            self.setItemWidget(item, b2)
            item.setSizeHint(b2.sizeHint())
            item.setFlags(item.flags() & ~Qt.ItemIsDragEnabled)
            self.save_schema()

    #
    def delete_button(self, position):
        ret = QMessageBox().question(self, '', "Are you sure?", QMessageBox.Yes | QMessageBox.No)

        if ret == QMessageBox.Yes:
            item = self.takeItem(position)
            del item
            self.save_schema()
            self.score_changed.emit(self, self.exam_id)

    def add_comment(self, position):
        # help me get text comment with dialog
        item = self.item(position)
        widget = self.itemWidget(item)
        text, ok = QInputDialog.getText(self, 'Add Comment', 'Comment:', text=widget.get_comment())
        if ok:
            item = self.item(position)
            widget = self.itemWidget(item)
            widget.setToolTip(text)
            widget.set_comment(text)

    def duplicate_button(self, position):
        item = self.item(position)
        widget = self.itemWidget(item)
        name, valid = QInputDialog.getText(self, 'Duplicate', 'Name:', text=widget.get_name() + "_copy")
        if valid:
            button = widget.__class__(name, **widget.get_config())
            button.score_changed.connect(self.button_clicked)
            item = QListWidgetItem()
            self.addItem(item)
            self.setItemWidget(item, button)
            item.setSizeHint(button.sizeHint())
            self.save_schema()

    def edit_rubric_config(self):
        dialog = RubricEditDialog(self.config)
        if dialog.exec():
            self.save_schema()

    def set_shortcut_color(self, position):
        item = self.item(position)
        widget = self.itemWidget(item)
        color = QColorDialog.getColor()
        if color.isValid():
            widget.set_color(color.name())
            self.save_schema()

    def remove_shortcut(self, position):
        self.takeItem(position)
        self.save_schema()

    # def is_done(self, exam_id):
    #     done = False
    #     grades_for_this_exam = self.rubric_grades_data.get(exam_id)
    #     if grades_for_this_exam is None:
    #         return False
    #
    #     for b in self.buttons(StepButton):  # type: Score
    #         this_button = grades_for_this_exam.get(b.get_name(), {})
    #         value = this_button.get("value", -1)
    #         done = done or value != -1
    #     for b in self.buttons(Text):
    #         this_button = grades_for_this_exam.get(b.get_name(), {})
    #         value = this_button.get("text", "")
    #         done = done or value != ""
    #
    #     return done
    #
    def store(self, exam_id):
        if self.scores.get(exam_id) is None:
            self.scores[exam_id] = {}

        for b in self.filter_buttons(StateButton):  # type: Score
            self.scores[exam_id][b.get_name()] = b.get_state()
            # b.clear_comment()

    def retrieve(self, exam_id):
        self.current_exam_id = exam_id
        rubric_data = self.scores.get(exam_id, {})
        if rubric_data is not None:
            for button in self.filter_buttons(StateButton):  # type: Score
                button.blockSignals(True)
                button.set_state(rubric_data.get(button.get_name(), {}))
                button.blockSignals(False)

    def startDrag(self, ev):
        selected = self.selectedIndexes()[0].row()
        item = self.item(selected)
        widget = self.itemWidget(item)
        qd = QDrag(self)
        qd.setMimeData(self.model().mimeData(self.selectedIndexes()))
        pm = QPixmap(400, 20)
        pm.fill(Qt.transparent)
        qp = QPainter(pm)
        qp.drawText(10, 15, widget.get_name())
        qd.setPixmap(pm)
        qd.exec(ev, Qt.MoveAction)
        del qp

    schemaChanged = pyqtSignal()

    def dropEvent(self, event):
        super().dropEvent(event)
        self.schemaChanged.emit()  # type: ignore

    def filter_buttons(self, kind=StepButton):
        buttons = []
        for i in range(self.count()):
            item = self.item(i)
            widget = self.itemWidget(item)
            if isinstance(widget, kind):
                buttons.append(widget)
        return buttons



class Code:
    TYPE_A = 0
    TYPE_P = 1
    TYPE_Q = 2
    TYPE_N = 3
    TYPE_O = 4

    def __init__(self, data, x, y, w, h, page=None, pdf_page=None):
        self.data = data
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.marked = False
        self.exam = None
        self.page = page
        self.pdf_page = pdf_page
        if self.data[0] == "O":
            self.date = int(self.data[1:7])
            self.exam = int(self.data[7:10])
            self.question = int(self.data[10:12])
            self.type = self.TYPE_O
        elif self.data[0] == "P":
            self.date = int(self.data[1:7])
            self.exam = int(self.data[7:10])
            self.page = int(self.data[10:12])
            self.type = self.TYPE_P
        elif self.data[0] == "Q":
            self.date = int(self.data[1:7])
            self.exam = int(self.data[7:10])
            self.page = int(self.data[10:12])
            self.type = self.TYPE_Q
        elif self.data[0] == "N":
            self.date = int(self.data[1:7])
            self.exam = int(self.data[7:10])
            self.number = int(self.data[10:12])
            self.type = self.TYPE_N
        else:
            self.date = int(self.data[0:6])
            self.exam = int(self.data[6:9])
            self.question = int(self.data[9:11])
            self.answer = int(self.data[11])
            if self.answer not in [1, 2, 3, 4]:
                permut = self.permut[self.question - 1]
                self.answer = int(permut.index(self.answer)) + 1
                self.data = self.data[:11] + str(self.answer)
            self.type = self.TYPE_A

    def set_marked(self, marked):
        self.marked = marked

    def set_page(self, page):
        self.page = page

    def set_pdf_page(self, page):
        self.pdf_page = page

    def set_pos(self, pos):
        self.x = pos[0]
        self.y = pos[1]

    def move(self, delta):
        self.x += delta[0]
        self.y += delta[1]

    def get_exam_id(self):
        return self.exam

    def get_date(self):
        return self.date

    def get_page(self):
        return self.page

    def get_type(self):
        return self.type

    def get_data(self):
        return self.data

    def get_pos(self):
        return self.x, self.y

    def get_size(self):
        return self.w, self.h

    def set_size(self, *args):
        self.w, self.h = args

    def scale(self, factor):
        self.x, self.y = self.x * factor, self.y * factor
        self.w, self.h = self.w * factor, self.h * factor

    def __repr__(self):
        if self.type == self.TYPE_A:
            return f"({self.data}, {self.exam}, {self.x}, {self.y}, {self.w}, {self.h}, PAG:{self.page}, Q:{self.question}, A:{self.answer}), M:{self.marked}"
        elif self.type in [self.TYPE_P, self.TYPE_Q]:
            return f"({self.data}, {self.exam}, {self.x}, {self.y}, {self.w}, {self.h}, PAG:{self.page})"
        elif self.type == self.TYPE_N:
            return f"({self.data}, {self.exam}, {self.x}, {self.y}, {self.w}, {self.h}, PAG:{self.page}, NUM:{self.number})"
        elif self.type == self.TYPE_O:
            return f"({self.data}, {self.exam}, {self.x}, {self.y}, {self.w}, {self.h}, PAG:{self.page}, Q:{self.question})"

    permut = [[5, 6, 7, 8],
              [5, 6, 8, 7],
              [5, 7, 6, 8],
              [5, 7, 8, 6],
              [5, 8, 6, 7],
              [5, 8, 7, 6],  # 6
              [6, 5, 7, 8],  # 7
              [6, 5, 8, 7],  # 8
              [6, 7, 5, 8],  # 9
              [6, 7, 8, 5],  # 10
              [6, 8, 5, 7],  # 11
              [6, 8, 7, 5],  # 12
              [7, 5, 6, 8],  # 13
              [7, 5, 8, 6],  # 14
              [7, 6, 5, 8],  # 15
              [7, 6, 8, 5],  # 16
              [7, 8, 5, 6],  # 17
              [7, 8, 6, 5],  # 18
              [8, 5, 6, 7],  # 19
              [8, 5, 7, 6],  # 20
              [8, 6, 5, 7],  # 21
              [8, 6, 7, 5],
              [8, 7, 5, 6],
              [8, 7, 6, 5],
              [5, 6, 7, 8],
              [5, 6, 8, 7],
              [5, 7, 6, 8],
              [5, 7, 8, 6],
              [5, 8, 6, 7],
              [5, 8, 7, 6],
              [6, 5, 7, 8],
              [6, 5, 8, 7],
              [6, 7, 5, 8],
              [6, 7, 8, 5],
              [6, 8, 5, 7],
              [6, 8, 7, 5],
              [7, 5, 6, 8],
              [7, 5, 8, 6],
              [7, 6, 5, 8],
              [7, 6, 8, 5],
              [7, 8, 5, 6],
              [7, 8, 6, 5],
              [8, 5, 6, 7],
              [8, 5, 7, 6],
              [8, 6, 5, 7],
              [8, 6, 7, 5],
              [8, 7, 5, 6],
              [8, 7, 6, 5]]



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QSizePolicy, QLabel, QVBoxLayout


class WidgetsRow(QWidget):
    def __init__(self, *args, **kwargs):
        super(WidgetsRow, self).__init__()
        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)
        self.widgets = args
        self.name = None
        for widget in args: # type: QWidget
            if type(widget) == str:
                self.name = widget if self.name is None else self.name
                widget = QLabel(widget)
            widget.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
            widget.setContentsMargins(0, 0, 0, 0)
            widget.setMaximumWidth(100)
            widget.setMinimumWidth(100)
            self.layout.addWidget(widget)
        self.name = kwargs.get("name", self.name)


class VBox(QVBoxLayout):
    def __init__(self):
        super(VBox, self).__init__()
        self.widgets = {}

    def addWidget(self, a0, *args, **kwargs):
        super().addWidget(a0, *args, **kwargs)
        self.setAlignment(a0, Qt.AlignTop)
        if hasattr(a0, 'name'):
            self.widgets[a0.name] = a0


from PyQt5 import QtWidgets, QtGui
from PyQt5.QtCore import QTimer
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import (QLineEdit, QPushButton, QSpinBox, QDialog, QDialogButtonBox, QComboBox, QDoubleSpinBox, QFormLayout, QCheckBox)

from qrgrader.widget_utils import WidgetsRow, VBox


class ButtonEditDialog(QDialog):
    def __init__(self, draggable_list, button=None):
        super().__init__()

        schema = button.get_config() if button is not None else {}

        self.draggable_list = draggable_list
        self.setWindowTitle("Edit")

        dialog_ok_cancel_btn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(dialog_ok_cancel_btn)
        self.buttonBox.accepted.connect(self.accept)  # type: ignore
        self.buttonBox.rejected.connect(self.reject)  # type: ignore
        self.buttonBox.buttons()[0].setEnabled(False)

        self.layout = VBox()
        self.setLayout(self.layout)

        self.combo = QComboBox()
        self.combo.addItems(['button', 'multiplier', 'cutter', 'text', 'separator'])
        self.combo.setCurrentText(schema.get('type', 'button'))
        self.layout.addWidget(WidgetsRow("Type", self.combo))

        self.le = QLineEdit()
        self.le.textChanged.connect(lambda: self.buttonBox.buttons()[0].setEnabled(self.le.text() != ""))  # type: ignore
        self.le.setText(button.get_name() if button is not None else "")
        self.layout.addWidget(WidgetsRow("Name", self.le))

        # self.value = QComboBox()
        self.value = QDoubleSpinBox()
        self.value.setDecimals(1)
        self.value.setMinimum(-20)
        self.value.setMaximum(20)
        self.value.setValue(float(schema.get('full_value', 1)))
        self.value.setSingleStep(0.5)
        self.value.valueChanged.connect(self.spin_value_changed)  # type: ignore

        # self.fill_cb(schema.get("type", 'button'), schema.get('value', 1))
        self.layout.addWidget(WidgetsRow("Value", self.value))

        self.steps = QSpinBox()
        self.steps.setMinimum(0)
        self.steps.setMaximum(10)
        self.steps.setValue(int(schema.get('steps', 0)))
        self.layout.addWidget(WidgetsRow("Steps", self.steps))
        #
        self.percent = QSpinBox()
        self.percent.setMinimum(0)
        self.percent.setMaximum(100)
        self.percent.setValue(int(schema.get('percent', 1) * 100))
        self.layout.addWidget(WidgetsRow("Percent", self.percent))
        #
        self.weight = QDoubleSpinBox()
        self.weight.setDecimals(1)
        self.weight.setMinimum(-20)
        self.weight.setMaximum(20)
        self.weight.setValue(float(schema.get('weight', 1)))
        self.weight.setSingleStep(0.5)
        self.layout.addWidget(WidgetsRow("Weight", self.weight))

        self.click_next_cb = QCheckBox("Next on click")
        self.click_next_cb.setChecked(schema.get('click_next', False))
        self.layout.addWidget(self.click_next_cb)

        # Show the current color and a color picker button
        self.color = QColor(schema.get('color', "#D4D4D4"))
        self.colorButton = QPushButton('Choose color')
        self.colorButton.clicked.connect(self.pick_color)  # type: ignore
        self.layout.addWidget(WidgetsRow('Color', self.colorButton))
        self.colorButton.setStyleSheet(f'background-color: {schema.get("color", "#D4D4D4")}')

        self.layout.addWidget(self.buttonBox)
        self.combo.currentTextChanged.connect(self.cb_changed)  # type: ignore

        self.enable_widgets()

    def spin_value_changed(self, value):
        if value < 0:
            self.weight.setStyleSheet('background-color: red')
            self.weight.setValue(0)
            QTimer.singleShot(1000, lambda: self.weight.setStyleSheet(''))

    def cb_changed(self, text):
        #        self.fill_cb(self.combo.currentText(), 1)
        self.enable_widgets()

    def enable_widgets(self):
        b = self.combo.currentText() in ['button']
        bm = self.combo.currentText() in ['button', 'multiplier']
        bmc = self.combo.currentText() in ['button', 'multiplier', 'cutter']
        mc = self.combo.currentText() in ['multiplier', 'cutter']

        self.layout.widgets['Value'].setVisible(b)
        self.layout.widgets['Steps'].setVisible(b)
        self.layout.widgets['Weight'].setVisible(b)
        self.layout.widgets['Color'].setVisible(bmc)
        self.layout.widgets['Percent'].setVisible(mc)
        self.click_next_cb.setVisible(b)

        self.adjustSize()

    def pick_color(self, button):
        color = QtWidgets.QColorDialog.getColor()
        if color.isValid():
            self.color = color
            self.colorButton.setStyleSheet(f'background-color: {color.name()}')

    def get_stylesheet(self):
        return f'background-color: {self.color.name()}'

    def get(self):
        res = {'type': self.combo.currentText()}
        if self.combo.currentText() in ['button', 'multiplier', 'cutter']:
            res['color'] = self.color.name()
        if self.combo.currentText() in ['multiplier', 'cutter']:
            res['percent'] = float(self.percent.value() / 100.0)
        if self.combo.currentText() in ['button']:
            res['steps'] = self.steps.value()
            res['weight'] = float(self.weight.value())
            res['full_value'] = float(self.value.value())
            res['click_next'] = self.click_next_cb.isChecked()

        return self.le.text(), self.combo.currentText(), res


class RubricEditDialog(QDialog):
    def __init__(self, config):
        super().__init__()
        self.config = config
        self.setWindowTitle("Edit")

        dialog_ok_cancel_btn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(dialog_ok_cancel_btn)
        self.buttonBox.accepted.connect(self.accept)  # type: ignore
        self.buttonBox.rejected.connect(self.reject)  # type: ignore
        self.buttonBox.buttons()[0].setEnabled(False)

        self.layout = QFormLayout()
        self.setLayout(self.layout)

        self.combo = QLineEdit()
        self.combo.setValidator(QtGui.QIntValidator(1, 99))
        self.layout.addRow("Page", self.combo)
        self.combo.setText(str(self.config.get("page", 1)))

        self.le = QLineEdit()

        # limit to 1 decimal
        self.le.setValidator(QtGui.QDoubleValidator(0, 10, 1))

        self.le.textChanged.connect(lambda: self.buttonBox.buttons()[0].setEnabled(self.le.text() != ""))  # type: ignore
        self.le.setText(str(self.config.get("weight", 10)))
        self.layout.addRow("Weight", self.le)

        self.layout.addWidget(self.buttonBox)

    # if accepted, modify the config
    def accept(self):
        self.config["weight"] = float(self.le.text())
        self.config["page"] = int(self.combo.text())
        super().accept()



import os

import pandas

from qrgrader.code import Code
from qrgrader.code_set import CodeSet


def get_workspace_paths(base):
    dir_workspace = base + os.sep
    dir_data = dir_workspace + "data" + os.sep
    dir_scanned = dir_workspace + "scanned" + os.sep
    dir_generated = dir_workspace + "generated" + os.sep
    dir_xls = dir_workspace + "results" + os.sep + "xls" + os.sep
    dir_publish = dir_workspace + "results" + os.sep + "pdf" + os.sep
    dir_source = dir_workspace + "source" + os.sep
    return dir_workspace, dir_data, dir_scanned, dir_generated, dir_xls, dir_publish, dir_source


def get_temp_paths(date, base):
    dir_scanner = base + os.sep + "__qrgrading__" + date + os.sep + "scanner" + os.sep
    dir_generator = base + os.sep + "__qrgrading__" + date + os.sep + "generator" + os.sep
    return dir_scanner, dir_generator


def get_date():
    if not check_workspace():
        raise Exception("get_date must be used from within a workspace directory")

    date = os.getcwd().split("-")[1].strip("/")
    return date


def check_workspace():
    current_dir = os.getcwd()
    dir_name = os.path.basename(current_dir)
    name = dir_name.split("-")
    if len(name) != 2 or name[0] != "qrgrading" or len(name[1]) != 6 or not name[1].isdigit():
        return False
    return True


def get_prefix():
    date = get_date()
    return str(date) + "_"


class Questions:
    def __init__(self, filename):
        self.filename = filename
        self.questions = None

    def load(self):
        if os.path.exists(self.filename):
            self.questions = pandas.read_csv(self.filename, sep='\t', header=0)
            return True
        return False

    def get_text(self, question):
        return self.questions.loc[question - 1, "BRIEF"]

    def get_value(self, question, answer):
        return self.questions.loc[question - 1, chr(answer + 64)]

    def get_type(self, question):
        return self.questions.loc[question - 1, "TYPE"]

    def get_questions(self):
        return self.questions.iloc[:, 0].tolist()


class Generated(CodeSet):

    def __init__(self, ppm):
        super().__init__()
        self.ppm = ppm

    def load(self, filename):
        if not os.path.exists(filename):
            return False

        with open(filename, "rb") as f:
            for line in f.readlines():
                data, x, y, w, h, pag, pdf_pag = line.decode().strip().split(",")
                x = int(int(x) / 65535 * 0.351459804 * self.ppm)
                y = int(297 * self.ppm - int(int(y) / 65535 * 0.351459804 * self.ppm))  # 297???
                self.append(Code(data, int(x), int(y), 120, 120, int(pag), int(pdf_pag)))
        return True


class StudentsData:
    def __init__(self, filename):
        self.filename = filename
        self.data = None

    def load(self):
        if os.path.exists(self.filename):
            self.data = pandas.read_csv(self.filename, sep='\t', header=0)
            return True
        return False

    def get_name(self, nia):
        if self.data is None:
            return None
        by_nia = self.data[self.data["NIA"] == nia]
        if by_nia.empty:
            return None

        return by_nia.iloc[0]["NAME"]

    def get_group(self, nia):
        if self.data is None:
            return None
        by_nia = self.data[self.data["NIA"] == nia]
        if by_nia.empty:
            return None

        return by_nia.iloc[0]["GROUP"]


class Nia:

    def __init__(self, filename):
        self.filename = filename
        self.nia = None

    def load(self):
        if os.path.exists(self.filename):
            self.nia = pandas.read_csv(self.filename, sep='\t', header=0)
            return True
        return False

    def get_nia(self, exam):
        by_exam = self.nia[self.nia["EXAM"] == int(exam)]
        if by_exam.empty or by_exam.iloc[0]["NIA"] is None:
            return None
        nia = by_exam.iloc[0]["NIA"]
        return int(nia) if str(nia).isdigit() else str(nia)

    def get_exam(self, nia):
        by_nia = self.nia[self.nia["NIA"] == nia]
        if by_nia.empty:
            return None

        return by_nia.iloc[0]["EXAM"]


def get_narrowest_type(cell):
    try:
        res = int(cell)
    except ValueError:
        try:
            res = float(cell)
        except ValueError:
            res = str(cell)

    return res



import argparse
import os
import sys
import time
from multiprocessing import Manager, Pool, Process
from random import randint

import cv2
import pymupdf
from PyQt5.QtCore import QObject, pyqtSignal
from PyQt5.QtWidgets import QGraphicsRectItem
from pymupdf.mupdf import PDF_ENCRYPT_KEEP

from qrgrader.code import Code
from qrgrader.code_set import CodeSet, PageCodeSet
from qrgrader.common import check_workspace, get_workspace_paths, get_temp_paths, Generated, Questions, get_date
from qrgrader.page_processor import PageProcessor
from qrgrader.utils import makedir


def main():
    parser = argparse.ArgumentParser(description='Patching and detection')

    parser.add_argument('-B', '--begin', type=int, help='First page to process', default=0)
    parser.add_argument('-E', '--end', type=int, help='Last page to process', default=None)
    parser.add_argument('-R', '--ratio', type=int, help='Resize image to save space', default=0.25)
    parser.add_argument('-s', '--scan', help='Process pages in scanned folder', action="store_true")
    parser.add_argument('-d', '--dpi', help='Dot per inch', type=int, default=400)
    parser.add_argument('-a', '--annotate', help='Annotate files', action="store_true")
    parser.add_argument('-n', '--nia', help='Create NIA file', action="store_true")
    parser.add_argument('-r', '--raw', help='Create RAW file', action="store_true")
    parser.add_argument('-S', '--simulate', help='Create random marked files', type=int, default=0)
    parser.add_argument('-e', '--reconstruct', help='Reconstruct exams', action="store_true")
    parser.add_argument('-p', '--process', help='Options -snre', action="store_true")
    parser.add_argument('-T', '--temp', help='Specify temp directory', type=str, default="/tmp")

    args = vars(parser.parse_args())

    if not check_workspace():
        print("ERROR: qrscanner must be run from a workspace directory")
        sys.exit(1)

    dir_workspace, dir_data, dir_scanned, dir_generated, dir_xls, dir_publish, dir_source = get_workspace_paths(os.getcwd())
    dir_temp_scanner, _ = get_temp_paths(get_date(), os.getcwd() if args.get("temp") is None else args.get("temp"))

    prefix = str(get_date()) + "_"

    ppm = args.get("dpi") / 25.4
    if args.get("process"):
        args["scan"] = True
        args["nia"] = True
        args["raw"] = True
        args["reconstruct"] = True

    if (simulate := args.get("simulate")) > 0:
        print("Simulation in progress ({} files)".format(simulate))
        makedir(dir_scanned, clear=True)

        generated = Generated(72 / 25.4)

        if not generated.load(dir_data + prefix + "generated.csv"):
            print("ERROR: generated.csv not found")
            sys.exit(1)

        pdf_filenames = [f for f in os.listdir(dir_generated) if f.endswith(".pdf")]
        pdf_filenames.sort()

        for pdf_filename in pdf_filenames[0:simulate]:
            print("Marking random answers in {}".format(pdf_filename), end="\r")

            new_pdf = pymupdf.open()
            doc = pymupdf.open(dir_generated + pdf_filename)

            exam = pdf_filename[6:9]

            for page in doc:

                filtered = generated.select(type=Code.TYPE_A, exam=int(exam), page=page.number + 1)
                for question in filtered.get_questions():
                    qrs = filtered.select(type=Code.TYPE_A, question=question)
                    for qr in qrs:
                        if qr.answer == randint(1, 4):
                            x = qr.x + 5
                            y = qr.y - 7
                            w = 10
                            h = 10
                            annot = page.add_redact_annot(pymupdf.Rect(x, y, x + w, y + h), fill=(0.5, 0.5, 0.5), cross_out=False)
                            break

                nias = generated.select(type=Code.TYPE_N, exam=int(exam), page=page.number + 1)
                for i in range(6):
                    r = randint(0, 9)
                    qr = nias.first(number=i * 10 + r)
                    if qr is not None:
                        x = qr.x + 5
                        y = qr.y - 7
                        w = 10
                        h = 10
                        annot = page.add_redact_annot(pymupdf.Rect(x, y, x + w, y + h), fill=(0.5, 0.5, 0.5), cross_out=False)

                # Apply the redactions
                page.apply_redactions()
                # Get the images with the marked codes
                pix = page.get_pixmap(matrix=pymupdf.Matrix(args.get("dpi") / 72, args.get("dpi") / 72))
                # Insert the new page and the image inside it
                new_page = new_pdf.new_page()
                new_page.insert_image(new_page.rect, stream=pix.tobytes("jpg"), width=pix.width, height=pix.height)
                # To be sure avoiding memory leaks
                del pix

            new_pdf.save(dir_scanned + pdf_filename)
            new_pdf.close()
            doc.close()

        print("\nSimulation done.")

    if args.get("scan", False):

        os.makedirs(dir_temp_scanner, exist_ok=True)

        generated = Generated(ppm)
        if not generated.load(dir_data + prefix + "generated.csv"):
            print("ERROR: generated.csv not found")
            sys.exit(1)

        with Manager() as manager:

            processes = []
            detected = manager.list()

            files = [x for x in os.listdir(dir_scanned) if x.endswith(".pdf")]
            for filename in files:
                print("Processing file {}".format(filename))
                document = pymupdf.open(dir_scanned + filename)
                last_page = args.get("end") if args.get("end") is not None else len(document)
                document.close()

                for i in range(args.get("begin"), last_page):

                    # We send the filename and open the document in the process for three reasons:
                    # 1. Sending the page object is not possible because it is not pickable
                    # 2. Rendering the page image in parallel make the whole process much faster
                    # 3. For some reason sending the image to the process creates memory overflow

                    process = PageProcessor(dir_scanned + filename, i, generated, detected, dir_images=dir_temp_scanner, resize=args.get("ratio"))
                    processes.append(process)
                    process.start()

                    while len([p for p in processes if p.is_alive()]) >= 4:
                        time.sleep(0.25)

            for process in processes:
                process.join()

            codes = CodeSet()
            codes.extend(detected)
            codes.save(dir_data + prefix + "detected.csv")

    if args.get("reconstruct") or args.get("nia") or args.get("raw") or args.get("annotate"):
        codes = CodeSet()
        if not codes.load(dir_data + prefix + "detected.csv"):
            print("ERROR: detected.csv not found")
            sys.exit(1)

        exams = codes.get_exams()
        date = codes.get_date()

    if args.get("reconstruct"):
        print("Reconstructing exams")
        images = os.listdir(dir_temp_scanner)
        for exam in exams:
            filename = dir_publish + "{}{:03d}.pdf".format(date, exam)
            pdf_file = pymupdf.open()
            exam_images = sorted([x for x in images if x.startswith("page-{}-{}-".format(date, exam))])
            for image in exam_images:
                page = pdf_file.new_page()  # noqa
                page.insert_image(pymupdf.Rect(0, 0, 595.28, 842), filename=dir_temp_scanner + os.sep + image)

            pdf_file.save(filename)

    if args.get("nia"):
        print("Creating NIA xls file")
        with open(dir_xls + prefix + "nia.csv", "w", encoding='utf-8') as f:
            f.write("EXAM\tNIA\n")
            for exam in exams:
                nia = {0: 'Y', 1: 'Y', 2: 'Y', 3: 'Y', 4: 'Y', 5: 'Y'}
                for row in range(6):
                    for number in range(10):
                        result = codes.first(exam=exam, type=Code.TYPE_N, number=row * 10 + number)
                        if result is None or result.marked:
                            nia[row] = number if nia[row] == 'Y' else 'X'
                nia = "".join([str(x) for x in nia.values()])

                f.write("{}\t{}\n".format(date * 1000 + exam, nia))

    if args.get("raw"):
        print("Creating RAW xls file")
        with open(dir_xls + prefix + "raw.csv", "w", encoding='utf-8') as f:
            # # Header
            # line = "DATE\tEXAM"
            # for qn in codes.get_questions():
            #     for an in codes.get_answers():
            #         line += "\tQ{:d}{}".format(qn, chr(64 + an))
            # for on in codes.get_open():
            #     line += "\tO{:d}".format(on)
            # f.write(line + "\n")

            # Exams
            for exam in exams:
                line = f"{date}\t{exam}"
                for question in codes.get_questions():
                    for answer in codes.get_answers():
                        result = codes.first(exam=exam, type=Code.TYPE_A, question=question, answer=answer)
                        line += "\t1" if result is None or result.marked else "\t0"

                for open_questions in codes.get_open():
                    line += "\t0"
                f.write(line + "\n")

    if args.get("annotate"):
        print("Annotating exams")
        questions = Questions(dir_xls + prefix + "questions.csv")
        if not questions.load():
            print("ERROR: questions.csv not found")
            sys.exit(1)

        for exam in exams:
            filename = dir_publish + "{}{:03d}.pdf".format(date, exam)
            pdf_file = pymupdf.open(filename)
            for page in pdf_file:

                for annot in page.annots():
                    page.delete_annot(annot)

                this_page = codes.select(type=Code.TYPE_A, exam=exam, page=page.number + 1)
                for code in this_page:
                    if code.marked:
                        x, y = code.get_pos()
                        w, h = code.get_size()
                        r = pymupdf.Rect(x, y, x + w, y + h)
                        annot = page.add_rect_annot(r)
                        annot.set_border(width=2)
                        if questions.get_value(code.question, code.answer) > 0:
                            annot.set_colors(stroke=(0, 1, 0))
                        else:
                            annot.set_colors(stroke=(1, 0, 0))
                        annot.update()
            pdf_file.save(filename, incremental=True, encryption=PDF_ENCRYPT_KEEP)

    print("All done :)")


if __name__ == '__main__':
    main()



from PyQt5.QtCore import QTimer, Qt
from PyQt5.QtWidgets import QTreeWidget, QHeaderView


class MyTreeHeader(QHeaderView):
    def __init__(self, orientation, parent=None):
        super().__init__(orientation, parent)

    def mousePressEvent(self, event):
        # Get the clicked column index
        column = self.logicalIndexAt(event.pos())
        if column != 0:
            super().mousePressEvent(event)


class PDFTree(QTreeWidget):

    def __init__(self):
        super().__init__()

        self.enabled = True
        self.setHeaderLabels(["#", "Exam Id", "!", "Score"])
        self.setSortingEnabled(True)
        self.setHeader(MyTreeHeader(Qt.Horizontal, self))
        self.header().setMinimumSectionSize(15)
        self.header().resizeSection(2, 15)  # Set initial width for column 0
        self.header().setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self.header().setSectionResizeMode(1, QHeaderView.ResizeToContents)
        self.header().setSectionResizeMode(3, QHeaderView.ResizeToContents)
        self.header().sortIndicatorChanged.connect(self.sort)
        self.setMaximumWidth(200)

    def sort(self, index, order):
        if index > 0:
            self.renumber()

    def renumber(self):
        j = 1
        for i in range(self.topLevelItemCount()):
            item = self.topLevelItem(i)
            if not item.isHidden():
                item.setText(0, str(j))
                j = j + 1

    def addTopLevelItem(self, item):
        super().addTopLevelItem(item)
        self.sortByColumn(1, Qt.AscendingOrder)
        self.renumber()

    def keyPressEvent(self, event):
        if self.enabled:
            super().keyPressEvent(event)
        else:
            event.ignore()

    def mousePressEvent(self, event):
        if self.enabled:
            super().mousePressEvent(event)
        else:
            event.ignore()

    def set_enabled(self, value):
        self.enabled = value



import os

import pandas
import pandas as pd

from qrgrader.common import Nia, get_workspace_paths, get_date, StudentsData, Questions


class Raw:
    def __init__(self, filename):
        self.filename = filename
        self.raw = None

    def load(self):
        if os.path.exists(self.filename):
            self.raw = pandas.read_csv(self.filename, sep='\t', header=None)
            return True
        return False

    def get_row(self, exam_id):
        if self.raw is None:
            return None
        by_exam = self.raw[self.raw[1] == exam_id]
        if by_exam.empty:
            return None
        return by_exam.iloc[0, :].tolist()

    def get_exams(self):
        if self.raw is None:
            return None
        return self.raw.iloc[:, 1].tolist()


def main():
    (dir_workspace,
     dir_data,
     dir_scanned, _,
     dir_xls,
     dir_publish, _) = get_workspace_paths(os.getcwd())

    date = get_date()

    nia = Nia(dir_xls + os.sep + str(date) + "_nia.csv")
    nia.load()

    students_data = StudentsData(dir_xls + os.sep + "DATA.csv")
    students_data.load()

    questions = Questions(dir_xls + os.sep + str(date) + "_questions.csv")
    questions.load()

    raw = pd.read_csv(dir_xls + os.sep + str(date) + "_raw.csv", sep='\t', header=None)
    raw.iloc[:, 0] = raw.iloc[:, 0] * 1000 + raw.iloc[:, 1]
    raw.set_index(raw.iloc[:, 0], inplace=True)

    hdr = 4  # Number of header rows
    inserted = 0

    # Insert the NIA column
    raw.insert(inserted + 2, "NIA", "")
    for exam_id in raw.iloc[:, 0].tolist():
        raw.loc[exam_id, "NIA"] = nia.get_nia(exam_id)
    inserted += 1

    # Insert the GRADE column
    raw.insert(inserted + 2, "GRADE", '=SUMPRODUCT(INDIRECT(ADDRESS(ROW(), COLUMN() + 1) & ":" '
                                      '& ADDRESS(ROW(), COLUMNS(A1:1))), INDIRECT(ADDRESS(4, COLUMN() + 1) & ":" '
                                      '& ADDRESS(4, COLUMNS(A1:1))))')
    inserted += 1

    # Fill the header
    percent_answ = '=SUM(OFFSET(INDIRECT("RC", FALSE), 1, 0, ROWS(A:A)-ROW(), 1))/COUNT(OFFSET(INDIRECT("RC", FALSE), 1, 0, ROWS(A:A)-ROW(), 1))'
    percent_ques = ('=SUM(OFFSET(INDIRECT("RC", FALSE), ' + str(hdr) + ', -3, ROWS(A:A) - ROW()-4, 4))/COUNT(OFFSET(INDIRECT("RC", FALSE), '
                    + str(hdr) + ', 0, ROWS(A:A)-ROW(), 1))')

    names, qn, ans_letter, ans_value, ans_perc = [""] * (inserted + 2), [""] * (inserted + 2), [""] * (inserted + 2), [""] * (inserted + 2), [""] * (
            inserted + 2)

    for question in questions.get_questions():
        if questions.get_type(question) == "Q":
            names.extend([questions.get_text(question), "", "", percent_ques])
            qn.extend([question, question, question, question])
            ans_value.extend([questions.get_value(question, i + 1) for i in range(4)])
            ans_letter.extend([chr(65 + i) for i in range(4)])
            ans_perc.extend([percent_answ] * 4)
        else:
            names.append(questions.get_text(question))
            qn.append(question)
            ans_letter.append("O")
            ans_value.append(questions.get_value(question, 1))
            ans_perc.append(percent_ques)

    raw.loc[-5] = names
    raw.loc[-4] = qn
    raw.loc[-3] = ans_letter
    raw.loc[-2] = ans_value
    raw.loc[-1] = ans_perc

    df = raw.sort_index().reset_index(drop=True)

    df.to_csv(dir_xls + os.sep + str(date) + "_table.csv", sep='\t', index=False, header=False)


def main2():
    (dir_workspace,
     dir_data,
     dir_scanned, _,
     dir_xls,
     dir_publish, _) = get_workspace_paths(os.getcwd())

    date = get_date()

    nia = Nia(dir_xls + os.sep + str(date) + "_nia.csv")
    nia.load()

    students_data = StudentsData(dir_xls + os.sep + "DATA.csv")
    students_data.load()

    questions = Questions(dir_xls + os.sep + str(date) + "_questions.csv")
    questions.load()

    raw = Raw(dir_xls + os.sep + str(date) + "_raw.csv")
    raw.load()

    fields = ["EXAM", "ID"]

    fields = fields + ["NIA", "NAM", "CQ"]

    for f in fields:
        print(f"{f}\t", end="")

    column = chr(ord("A") + len(fields))

    for i in questions.get_questions():
        kind = questions.get_type(i)
        if kind == "Q":
            for j in range(4):
                if j == 0:
                    brief = questions.get_text(i)
                    print(f"{brief}\t", end="")
                elif j == 3:
                    begin_column = chr(ord(column) - 3)
                    print(f"=sum({begin_column}6:{column})/count({begin_column}6:{column})\t", end="")
                else:
                    print(f"  \t", end="")
                column = chr(ord(column) + 1)
        else:
            print(f"{i}\t")

    for f in fields:
        print(f"{f}\t", end="")

    for i in questions.get_questions():
        kind = questions.get_type(i)
        if kind == "Q":
            for j in range(4):
                print(f"{i}\t", end="")
        else:
            print(f"{i}\t")

    for f in fields:
        print(f"{f}\t", end="")

    for i in questions.get_questions():
        kind = questions.get_type(i)
        if kind == "Q":
            for j in range(4):
                print(f"{chr(65 + j)}\t", end="")
        else:
            print(f"O\t")
    for f in fields:
        print(f"{f}\t", end="")
    for i in questions.get_questions():
        kind = questions.get_type(i)
        if kind == "Q":
            for j in range(4):
                print(f"{questions.get_value(i, j + 1)}\t", end="")
        else:
            print(f"O\t")

    for f in fields:
        print(f"{f}\t", end="")

    column = chr(ord("A") + len(fields))

    for i in questions.get_questions():
        kind = questions.get_type(i)
        if kind == "Q":
            for j in range(4):
                print(f"=sum({column}6:{column})/count({column}6:{column})\t", end="")
                column = chr(ord(column) + 1)
        else:
            print(f"O\t")

    line = 4

    for i in raw.get_exams():
        row = raw.get_row(i)
        print(f"{row.pop(0)}\t{row.pop(0)}\t", end="")

        for f in fields[2:]:
            if f == "CQ":
                column = chr(ord("A") + len(fields))

                print(f"=sumproduct({column}4:4,{column}{line}:{line})\t", end="")
            else:

                print(f"   \t", end="")

        for elem in row:
            print(f"{elem}\t", end="")
        print()
        line += 1


if __name__ == "__main__":
    main()



import math
import os
import shutil
from pathlib import Path

import cv2
import numpy as np
import zxingcpp


def pix2np(pix):
    im = np.frombuffer(pix.samples, dtype=np.uint8).reshape(pix.h, pix.w, pix.n)
    im = np.ascontiguousarray(im[..., [2, 1, 0]])  # rgb to bgr
    return im


def threshold(orig, th):
    if th == 0:
        img = orig.copy()
    else:
        ret, img = cv2.threshold(orig, 255 * float(th) / 100, 255, cv2.THRESH_BINARY)

    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    ret, thresh = cv2.threshold(gray, 200, 255, 1)

    return thresh


def get_patches(orig, ppm, size_mm, tolerance=0.25):
    # List for the patches
    patches = []

    # Find patches
    contours, h = cv2.findContours(orig, 1, 2)

    # Get other patches
    for cnt in contours:

        leng = cv2.arcLength(cnt, True)

        if leng < 0:
            continue

        approx = cv2.approxPolyDP(cnt, 0.01 * cv2.arcLength(cnt, True), True)

        if (len(approx) > 8) or (len(approx) < 4):
            continue

        x, y, w, h = cv2.boundingRect(cnt)

        if (w > ppm * size_mm * (1 - tolerance)) and (h > ppm * size_mm * (1 - tolerance)) and (
                w < ppm * size_mm * (1 + tolerance)) and (
                h < ppm * size_mm * (1 + tolerance)):  # and (y - ppm > 0) and (x - ppm > 0):
            patches.append((x, y, w, h))
        else:
            if (w > ppm * size_mm * 0.2) and (h > ppm * size_mm * 0.2) \
                    and (w < ppm * size_mm * 0.5) and (h < ppm * size_mm * 0.5):
                a = int(ppm * size_mm / 2)
                patches.append((x - a, y - a, w + 2 * a, h + 2 * a))

    return patches


def get_codes(patch):
    codes = set()
    results = zxingcpp.read_barcodes(patch, formats=zxingcpp.BarcodeFormat.Aztec | zxingcpp.BarcodeFormat.QRCode)
    for result in results:
        if result.text not in codes:
            top_left_x = min(result.position.top_left.x, result.position.bottom_right.x)
            top_left_y = min(result.position.top_left.y, result.position.bottom_right.y)
            width = abs(result.position.top_left.x - result.position.bottom_right.x)
            height = abs(result.position.top_left.y - result.position.bottom_right.y)
            codes.add((result.text, top_left_x, top_left_y, width, height))
    return codes


def compute_similarity_transform(p11, p12, p21, p22):
    # Convert points to numpy arrays
    p11, p12, p21, p22 = map(np.array, [p11, p12, p21, p22])

    # Compute centroids
    C1 = (p11 + p12) / 2.0
    C2 = (p21 + p22) / 2.0

    # Compute translation vector
    T = C2 - C1

    # Compute scale factor
    d1 = np.linalg.norm(p12 - p11)  # Distance in reference frame A
    d2 = np.linalg.norm(p22 - p21)  # Distance in reference frame B
    s = d2 / d1 if d1 != 0 else 1  # Avoid division by zero

    # Compute rotation angle
    delta1 = p12 - p11
    delta2 = p22 - p21

    theta1 = math.atan2(delta1[1], delta1[0])
    theta2 = math.atan2(delta2[1], delta2[0])
    theta = theta2 - theta1  # Rotation angle in radians

    # Construct rotation matrix
    R = np.array([[math.cos(theta), -math.sin(theta)],
                  [math.sin(theta), math.cos(theta)]])

    return s, R, T


def makedir(path, clear=False):
    try:
        os.makedirs(path)
    except FileExistsError:
        if clear:
            directory = Path(path)
            for item in directory.glob('*'):
                if item.is_file() or item.is_symlink():
                    item.unlink()



import argparse
import os
import sys
import time
from multiprocessing import Manager, Pool, Process

import cv2
import pymupdf

from qrgrader.code import Code
from qrgrader.code_set import CodeSet, PageCodeSet
from qrgrader.utils import pix2np, get_patches, threshold, get_codes, compute_similarity_transform


class PageProcessor(Process):

    def __init__(self, filename, index, generated, result, **kwargs):
        super().__init__()
        self.filename = filename
        self.index = index
        self.generated = generated
        self.result = result

        self.dpi = kwargs.get("dpi", 400)
        self.thresholds = kwargs.get("thresholds", [50, 55, 60, 65, 70, 75, 80])
        self.matrix = pymupdf.Matrix(self.dpi / 72, self.dpi / 72)
        self.show_patches = kwargs.get("show_patches", False)
        self.resize = kwargs.get("resize", 1.0)
        self.dir_images = kwargs.get("dir_images", "../..")

        self.ppm = self.dpi / 25.4

    def run(self):

        # Render the image
        doc = pymupdf.open(self.filename)
        page = doc[self.index]
        image = pix2np(page.get_pixmap(matrix=self.matrix))  # noqa
        doc.close()

        # Find page, orientation and rotate page
        rotation = None
        detected = PageCodeSet()

        for th in self.thresholds:
            ph, pw = image.shape[0], image.shape[1]

            nw = image[0:500, 0:500]
            ne = image[0:500, pw - 500:pw]
            sw = image[ph - 500:ph, 0:500]
            se = image[ph - 500:ph, pw - 500:pw]

            nw = threshold(nw, th)
            ne = threshold(ne, th)
            sw = threshold(sw, th)
            se = threshold(se, th)

            # cv2.imwrite("nw.png", nw)
            # cv2.imwrite("sw.png", sw)
            # cv2.imwrite("ne.png", ne)
            # cv2.imwrite("se.png", se)

            for text, cx, cy, cw, ch in get_codes(ne):
                if text.startswith("P"):
                    rotation = None
                elif text.startswith("Q"):
                    rotation = cv2.ROTATE_90_CLOCKWISE
                detected.append(Code(text, cx, cy, cw, ch))

            for text, cx, cy, cw, ch in get_codes(nw):
                if text.startswith("P"):
                    rotation = cv2.ROTATE_90_CLOCKWISE
                elif text.startswith("Q"):
                    rotation = cv2.ROTATE_90_COUNTERCLOCKWISE
                detected.append(Code(text, cx, cy, cw, ch))

            for text, cx, cy, cw, ch in get_codes(sw):
                if text.startswith("P"):
                    rotation = cv2.ROTATE_90_COUNTERCLOCKWISE
                elif text.startswith("Q"):
                    rotation = None
                detected.append(Code(text, cx, cy, cw, ch))

            for text, cx, cy, cw, ch in get_codes(se):
                if text.startswith("P"):
                    rotation = cv2.ROTATE_180
                elif text.startswith("Q"):
                    rotation = cv2.ROTATE_90_COUNTERCLOCKWISE
                detected.append(Code(text, cx, cy, cw, ch))

        if rotation is not None:
            image = cv2.rotate(image, rotation)

        # Get the page number if we got it
        page = detected.get_page() if detected.get_page() is not None else 0

        # Clear the detected because the
        # page may have been rotated
        detected.clear()

        # Process the page and extract the detected
        for th in self.thresholds:
            th_image = threshold(image, th)
            patches = get_patches(th_image, self.ppm, 8)

            for px, py, pw, ph in patches:
                patch = image[py:py + ph, px:px + pw]

                if self.show_patches:
                    cv2.rectangle(image, (px, py), (px + pw, py + ph), (0, 0, 255), 1)

                for text, cx, cy, cw, ch in get_codes(patch):
                    detected.append(Code(text, px + cx, py + cy, cw, ch, page, self.index))

        # Try again with the whole page
        page = detected.get_page()
        exam = detected.get_exam_id()

        # If we did not find the page number, try to find it in the generated detected
        if page is None:
            for code in detected:
                code = self.generated.get(code)
                if code := self.generated.get(code):
                    page = code.get_page()
                    exam = code.get_exam_id()
                    break
            for code in detected:
                code.set_page(page)

        if self.resize != 1.0:
            image = cv2.resize(image, (int(image.shape[1] * self.resize), int(image.shape[0] * self.resize)),
                               interpolation=cv2.INTER_AREA)

        if detected.get_page() is not None:
            cv2.imwrite(self.dir_images + os.sep + "page-{}-{}-{:03d}.jpg".format(detected.get_date(), detected.get_exam_id(), detected.get_page()), image)
        elif detected.get_exam_id():
            cv2.imwrite(self.dir_images + os.sep + "page-{}-{}-{:03d}.jpg".format(detected.get_date(), detected.get_exam_id(), 0), image)
        else:
            cv2.imwrite(self.dir_images + os.sep + "{}-{:03d}.jpg".format(self.filename, self.index), image)

        # # Compute the transformation
        generated_page_codeset = PageCodeSet(self.generated.select(exam=exam, page=page))

        # Compute the transformation
        if detected.get_p() is not None and detected.get_q() is not None:
            p11 = detected.get_p().get_pos()
            p12 = detected.get_q().get_pos()
            p21 = generated_page_codeset.get_p().get_pos()
            p22 = generated_page_codeset.get_q().get_pos()
            _, _, delta = compute_similarity_transform(p21, p22, p11, p12)
        elif detected.get_p() is not None:
            x1, y1 = detected.get_p().get_pos()
            x2, y2 = generated_page_codeset.get_p().get_pos()
            delta = (x2 - x1, y2 - y1)
        elif detected.get_q() is not None:
            x1, y1 = detected.get_q().get_pos()
            x2, y2 = generated_page_codeset.get_q().get_pos()
            delta = (x2 - x1, y2 - y1)
        else:
            delta = (0, 0)

        for code in generated_page_codeset:
            code.move(delta)
            code.set_size(120, 120)
            code.scale(72.0 / self.dpi)
            code.set_marked(detected.get(code) is None)
            self.result.append(code)

        print("Processed", self.filename, "page", self.index, len(self.result))



#!/usr/bin/env python

from __future__ import print_function

import argparse
import os.path
import re
import sys
from multiprocessing import Manager
from os import listdir

from qrgrader.common import get_workspace_paths, get_temp_paths, check_workspace, get_date, get_prefix
from qrgrader.generator import Generator
from qrgrader.utils import makedir


def main():
    parser = argparse.ArgumentParser(description='Patching and detection')
    parser.add_argument('-b', '--begin', help='First exam number', default=1, type=int)
    parser.add_argument('-c', '--cleanup', help='Clear pool files', action='store_true')
    parser.add_argument('-f', '--filename', help='Specify .tex filename', default=None)
    parser.add_argument('-j', '--threads', help='Maximum number of threads to use (4)', default=4, type=int)
    parser.add_argument('-n', '--number', help='Number or exams to be generated', default=0, type=int)
    parser.add_argument('-P', '--pages', help='Acceptable number of pages of output PDF', default=-1, type=int)
    parser.add_argument('-T', '--temp', help='Specify temp directory', type=str, default="/tmp")
    parser.add_argument('-q', '--questions', help='Generate questions csv', action='store_true')
    parser.add_argument('-g', '--generated', help='Generate questions csv', action='store_true')
    parser.add_argument('-v', '--verbose', help='Extra verbosity', action='store_true')
    args = vars(parser.parse_args())

    if not check_workspace():
        print("ERROR: qrscanner must be run from a workspace directory")
        sys.exit(1)

    date = get_date()
    prefix = get_prefix()

    (file, number, begin,
     threads, verbose, desired_pages,
     create_generated, create_questions, cleanup, temporary, filename) = (args["filename"], args["number"],
                                                                          args["begin"], args["threads"],
                                                                          args["verbose"], args["pages"],
                                                                          args["generated"], args["questions"],
                                                                          args["cleanup"], args["temp"], args["filename"])
    if number > 0:
        create_generated = True
        create_questions = True

    dir_workspace, dir_data, _, dir_generated, dir_xls, _, dir_source = get_workspace_paths(os.getcwd())
    _, dir_temp_generator = get_temp_paths(date, os.getcwd() if temporary is None else temporary)

    print("Workspace: {}".format(dir_temp_generator))

    if desired_pages != -1:
        print("WARNING: With -P option qrgenerator will generate a maximum of {} exams to get {} with {} pages".format(4 * number, number, desired_pages))
        print("WARNING: With -P option qrgenerator may generate up to {} more exams than needed.".format(threads))

    if number > 0:

        makedir(dir_generated, clear=True)
        makedir(dir_temp_generator, clear=True)

        if filename is None:
            source = dir_source
            filename = dir_source + "main.tex"
        else:
            source = os.path.dirname(os.path.abspath(filename))
            filename = os.path.basename(filename)

        if not os.path.exists(filename):
            print("ERROR: Source file {} does not exist".format(source + os.sep + filename))
            sys.exit(1)

        print("** Starting parallelized generation (using {} threads)".format(threads))

        end = begin + (number if desired_pages == -1 else 4 * number)

        processes = []

        with Manager() as manager:
            queue = manager.BoundedSemaphore(threads)

            for i in range(begin, end):
                queue.acquire()
                print("Creating exam {}{:03d} ({:d} ready)".format(date, i, len(os.listdir(dir_generated))))
                p = Generator(queue, filename, "{}{:03d}".format(date, i),
                              dir_temp_generator=dir_temp_generator,
                              dir_generated=dir_generated,
                              dir_source=source,
                              desired_pages=desired_pages,
                              verbose=verbose)

                processes.append(p)
                p.start()

            for p in processes:
                p.join()

            print("Done ({:d} exams generated).".format(len(os.listdir(dir_generated))))

    if create_generated:
        print("Creating generated.csv file...", end="")
        logs = sorted([x for x in listdir(dir_temp_generator) if x.endswith(".aux")])

        w = open(dir_data + prefix + "generated.csv", "w", encoding='utf-8')
        for f in logs:
            pdf_page = 0
            f = open(dir_temp_generator + os.sep + f, "r", encoding='utf-8')

            for line in f:
                if line.startswith("\\zref@newlabel{QRPOSE"):
                    words_with_numbers = re.findall(r'\b\w*\d\w*\b', line)
                    if len(words_with_numbers) != 5:
                        print("ERROR: QRPOSE line does not have 5 values: {}".format(line))
                        sys.exit(0)
                    else:
                        qr_data, posx, posy, abs_page, page_value = words_with_numbers

                        if qr_data.startswith("P"):
                            pdf_page += 1

                        line = "{},{},{},{},{},{},{}\n".format(qr_data, posx, posy, 0, 0, abs_page, pdf_page)
                        w.write(line)
            f.close()
        w.close()
        print("Done.")

    if create_questions:
        logs = [x for x in listdir(dir_temp_generator) if x.endswith(".log")]

        if len(logs) > 0:
            print("Creating questions csv file...", end="")

            any_log_will_do = dir_temp_generator + os.sep + logs[0]

            with open(any_log_will_do, "r", encoding='utf-8') as file:
                log = file.read()
                log = log.split(";;;")
                questions_number = []
                with open(dir_xls + prefix + "questions.csv".format(date), "w", encoding='utf-8') as filew:
                    data = "ID\tTYPE\tA\tB\tC\tD\tBRIEF\n"

                    for i in range(1, len(log) - 1):
                        line = log[i].replace("\n", "").replace("\t", "")
                        fields = line.split(";;")
                        fields = [f for f in fields if f != ""]
                        if len(fields) == 6:

                            if fields[0] in questions_number:
                                print("\n****************************************************")
                                print("*** WARNING: Multiple questions with same id: {:2s} ***".format(fields[0]))
                                print("****************************************************")
                            else:
                                questions_number.append(fields[0])

                            line = fields[0] + "\t" + fields[1] + \
                                   "\t" + (fields[3] if fields[2] == "a" else fields[4].replace(" ", "")) + \
                                   "\t" + (fields[3] if fields[2] == "b" else fields[4].replace(" ", "")) + \
                                   "\t" + (fields[3] if fields[2] == "c" else fields[4].replace(" ", "")) + \
                                   "\t" + (fields[3] if fields[2] == "d" else fields[4].replace(" ", "")) + \
                                   "\t" + fields[5]
                            data += line + "\n"
                    filew.write(data)

    print("Done.")


if __name__ == "__main__":
    main()



from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtGui import QCursor, QGuiApplication, QFont
from PyQt5.QtWidgets import QWidget, QPushButton, QHBoxLayout, QSpinBox, QLabel, QTextEdit, QVBoxLayout


class Button(QWidget):
    def __init__(self, name):
        super().__init__()
        self.name = name

    def get_name(self):
        return self.name


class Shortcut(Button):
    clicked = pyqtSignal()

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.button = QPushButton(name)
        self.button.clicked.connect(self.clicked.emit)  # type: ignore
        self.buttons = kwargs.get("buttons", [])
        self.set_color(kwargs.get("color", "#FF0000"))
        self.setLayout(QHBoxLayout())
        self.layout().addWidget(self.button)
        font = self.button.font()
        font.setItalic(True)
        self.button.setFont(font)

    def set_color(self, color):
        self.setStyleSheet('background-color:' + color)

    def get_buttons(self):
        return self.buttons

    def get_config(self):
        return {"type": "shortcut", "buttons": self.buttons, "color": self.styleSheet().split(":")[1]}


class Separator(Button):
    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.setLayout(QHBoxLayout())
        self.label = QLabel(name)
        self.label.setFont(QFont("Arial", 10, QFont.Bold))
        self.label.setAlignment(Qt.AlignCenter)
        self.layout().addWidget(self.label)
        self.layout().setContentsMargins(5, 5, 5, 5)

    def get_config(self):
        return {"type": "separator"}


class StateButton(Button):
    pass


class TextButton(StateButton):
    clicked = pyqtSignal()

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.button = QTextEdit(name)
        self.button.setMinimumWidth(10)
        self.setMinimumWidth(10)
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(QLabel(name))
        self.layout().addWidget(self.button)
        self.layout().setContentsMargins(5, 2, 5, 2)

    def get_config(self):
        return {"type": "text"}

    def get_state(self):
        return {"text": self.button.toPlainText()}

    def set_state(self, state):
        self.button.setText(state.get("text", ""))


class StepButton(StateButton):
    score_changed = pyqtSignal()

    def __init__(self, name, **kwargs):
        super().__init__(name)

        self.weight = kwargs.get('weight', 1)
        self.full_value = kwargs.get('full_value', 1)
        self.n_steps = kwargs.get('steps', 0)
        color = kwargs.get("color", "#D4D4D4")
        self.click_next = kwargs.get("click_next", False)

        self.comment = None
        self.step = int(100 / self.n_steps) if self.n_steps > 0 else 100

        layout = QHBoxLayout()
        layout.setContentsMargins(5, 2, 5, 2)

        self.button = QPushButton(name)
        self.button.setMinimumWidth(10)
        self.button.setCheckable(True)
        if color is not None:
            self.button.setStyleSheet('background-color: {}'.format(color))
        self.button.clicked.connect(self.clicked)

        self.spinner = QSpinBox()
        self.spinner.setEnabled(False)
        self.spinner.lineEdit().setReadOnly(True)
        self.spinner.setMinimum(0)
        self.spinner.setSingleStep(self.step)
        self.spinner.setMaximum(100)
        self.spinner.setMaximumWidth(50)
        self.spinner.valueChanged.connect(self.score_changed.emit)  # type: ignore

        layout.addWidget(self.button)
        self.comment_lbl = QLabel(self.button)
        self.comment_lbl.setGeometry(5, 5, 15, 15)
        self.comment_lbl.setText("*")
        self.comment_lbl.setStyleSheet('background-color: #00FFFF00;')
        self.comment_lbl.setMaximumWidth(10)
        self.comment_lbl.setVisible(False)

        self.set_comment(kwargs.get('comment', ""))
        self.start_with = kwargs.get('start_with', 100)

        if self.n_steps > 0:
            layout.addWidget(self.spinner)

        self.setLayout(layout)

    def get_color(self):
        return self.button.styleSheet().split(":")[1].strip()

    def get_config(self):
        return {"type": "button", "steps": self.n_steps, "weight": self.weight, "full_value": self.full_value,
                "color": self.get_color(), "comment": self.comment, "start_with": self.start_with,
                "click_next": self.click_next}

    # kwargs = {"weight": 1, "full_value": 1,
    #                "steps": 0, "color": None,
    #                "comment": "", "start_with": 100,
    #                "click_next": False, "type": "button"}

    def toggle_show_points(self):
        self.points_lb.setVisible(not self.points_lb.isVisible())

    def is_checked(self):
        return self.button.isChecked()

    def get_state(self):
        if self.is_checked():
            if self.n_steps == 0:
                value = 100
            else:
                value = self.spinner.value()
        else:
            value = -1

        return {"value": value, "comment": self.comment}

    def get_xls_value(self):
        state = self.get_state()
        value = state.get("value")
        if value == -1:
            return " "
        else:
            return value / 100

    def get_value(self):
        if self.is_checked():
            if self.n_steps == 0:
                return self.get_full_value()
            else:
                return float(self.spinner.value() / 100.0) * self.get_full_value()
        else:
            return None

    def clicked(self):
        self.spinner.blockSignals(True)
        if QGuiApplication.queryKeyboardModifiers() == Qt.ControlModifier:
            self.spinner.setValue(0)
        else:
            self.spinner.setValue(self.start_with if self.button.isChecked() else 0)
        self.spinner.setEnabled(self.button.isChecked())
        self.spinner.blockSignals(False)

        font = self.button.font()
        font.setBold(self.button.isChecked())
        self.button.setFont(font)

        self.score_changed.emit()  # type: ignore

    def set_state(self, state):
        self.set_comment(state.get("comment", ""))
        value = state.get("value", 0)

        self.button.blockSignals(True)
        self.button.setChecked(value >= 0 if self.n_steps > 0 else value > 0)
        self.button.blockSignals(False)

        self.spinner.blockSignals(True)
        self.spinner.setMinimum(0)
        self.spinner.setValue(value if value > 0 else 0)
        self.spinner.setEnabled(self.button.isChecked())
        self.spinner.blockSignals(False)

        font = self.button.font()
        font.setBold(self.button.isChecked())
        self.button.setFont(font)

        self.score_changed.emit()  # type: ignore

    def set_comment(self, text):
        self.comment_lbl.setVisible(text != "")
        self.setToolTip(text)
        self.comment = text

    def get_comment(self):
        return self.comment

    def clear_comment(self):
        self.set_comment("")

    def get_full_value(self):
        return self.full_value

    def click(self):
        self.button.click()

    def get_click_next(self):
        return self.click_next and self.button.isChecked()


class PercentButton(StateButton):
    score_changed = pyqtSignal()

    def __init__(self, name, **kwargs):
        super().__init__(name)

        self.percent = kwargs.get('percent', 1)
        color = kwargs.get("color", "#D4D4D4")
        self.setLayout(QHBoxLayout())
        self.layout().setContentsMargins(5, 2, 5, 2)
        self.button = QPushButton(name)
        self.button.setMinimumWidth(10)
        self.button.setCheckable(True)
        self.button.setStyleSheet('background-color: {}'.format(color))
        self.button.clicked.connect(self.score_changed.emit)
        self.layout().addWidget(self.button)

    def get_config(self):
        return {"percent": self.percent, "color": self.get_color()}

    def get_state(self):
        return {"percent": self.percent if self.button.isChecked() else 1}

    def set_state(self, state):
        self.button.setChecked(state.get("percent", 1) < 1)

    def get_xls_value(self):
        return self.percent if self.is_checked() else 1

    def get_color(self):
        return self.button.styleSheet().split(":")[1].strip()

    def get_percent(self):
        return self.percent if self.button.isChecked() else 1


class MultiplierButton(PercentButton):

    def get_config(self):
        config = super().get_config()
        config["type"] = "multiplier"
        return config


class CutterButton(PercentButton):

    def get_config(self):
        config = super().get_config()
        config["type"] = "cutter"
        return config



import os

from qrgrader.code import Code


class CodeSet:
    def __init__(self):
        self.codes = {}

    def append(self, code):
        self.codes[code.data] = code

    def extend(self, codes):
        for code in codes:
            self.append(code)

    def clear(self):
        self.codes.clear()

    def __repr__(self):
        text = str()
        for code in self.codes.values():
            text += str(code) + "\n"
        return text

    def __len__(self):
        return len(self.codes)

    def __next__(self):
        return next(iter(self.codes.values()))

    def __iter__(self):
        return iter(self.codes.values())

    def select(self, **kwargs):
        filtered = [x for x in self.codes.values() if all(getattr(x, key) == value for key, value in kwargs.items())]
        result = CodeSet()
        for code in filtered:
            result.append(code)
        return result

    def get(self, code):
        return self.codes.get(code.data)

    def get_exams(self):
        return sorted(list(set([x.exam for x in self.codes.values()])))

    def get_questions(self):
        return sorted(list(set([x.question for x in self.codes.values() if x.type == Code.TYPE_A])))

    def get_open(self):
        return sorted(list(set([x.question for x in self.codes.values() if x.type == Code.TYPE_O])))

    def get_answers(self):
        return sorted(list(set([x.answer for x in self.codes.values() if x.type == Code.TYPE_A])))

    def save(self, file_name):
        with open(file_name, "w", encoding='utf-8') as f:
            for code in self.codes.values():
                f.write(
                    code.data + ",{:.2f},{:.2f},{:.2f},{:.2f},{},{},{:d}\n".format(code.x, code.y, code.w, code.h, code.page, code.pdf_page, int(code.marked)))

    def load(self, file_name):
        if not os.path.exists(file_name):
            return False

        with open(file_name, "r", encoding='utf-8') as f:
            for line in f:
                fields = line.strip().split(",")
                data, x, y, w, h, page, pdf_page = fields[:7]
                code = Code(data, float(x), float(y), float(w), float(h), int(page), int(pdf_page))
                if len(fields) > 7:
                    code.set_marked(int(fields[7]))
                self.append(code)
        return True

    def get_date(self):
        if len(self.codes.values()) > 0:
            return next(iter(self.codes.values())).date
        return None

    def empty(self):
        return len(self) == 0

    def first(self, **kwargs) -> Code:
        return next(iter(self.select(**kwargs)), None)


class PageCodeSet(CodeSet):

    def __init__(self, codeset=None):
        super().__init__()
        if codeset is not None:
            self.codes = codeset.codes

    def get_q(self):
        return next((x for x in self.codes.values() if x.type == Code.TYPE_Q), None)

    def get_p(self):
        return next((x for x in self.codes.values() if x.type == Code.TYPE_P), None)

    def get_page(self):
        source = self.get_p() or self.get_q()
        if source is not None:
            return source.page
        return None

    def get_exam_id(self):
        if len(self.codes.values()) > 0:
            return next(iter(self.codes.values())).exam
        return None

    def get_date(self):
        if len(self.codes.values()) > 0:
            return next(iter(self.codes.values())).date
        return None



import argparse
import os
from datetime import date

from qrgrader.common import get_workspace_paths

import importlib.resources


def get_resource(name):
    with importlib.resources.files("qrgrader").joinpath("latex" + os.sep + name).open("r", encoding='utf-8') as f:
        return f.read()


def main():
    parser = argparse.ArgumentParser()
    today = date.today().strftime("%y%m%d")
    parser.add_argument('-d', '--date', type=int, help='Date', default=today)
    args = vars(parser.parse_args())

    if args["date"] < 100000 or args["date"] > 999999:
        print("Invalid date value, exiting.")
        exit(1)

    directories = get_workspace_paths(os.getcwd() + os.sep + "qrgrading-" + str(args["date"]))

    for directory in directories:
        if not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)

    dir_workspace, dir_data, dir_scanned, dir_generated, dir_xls, dir_publish, dir_source = directories

    with open(dir_source + "main.tex", "w", encoding='utf-8') as f:
        f.write(get_resource("main.tex"))

    with open(dir_source + "qrgrader.sty", "w", encoding='utf-8') as f:
        f.write(get_resource("qrgrader.sty"))

    print(f"Workspace qrgrader-{args['date']} created successfully.")


if __name__ == '__main__':
    main()



import os

import gspread
from gspread.utils import a1_to_rowcol, ValueInputOption
from pydrive2.auth import GoogleAuth
from pydrive2.drive import GoogleDrive

from qrgrader.common import get_narrowest_type


class GDrive:

    def __init__(self, config_dir=".", **kwargs):
        self.gdrive = None
        self.config_dir = config_dir
        if kwargs.get("authorize", True):
            self.authorize()

    def authorize(self):

        GoogleAuth.DEFAULT_SETTINGS['client_config_file'] = self.config_dir + os.sep + "client_secret.json"
        credentials_path = self.config_dir + os.sep + "credentials.json"

        gauth = GoogleAuth()
        gauth.LoadCredentialsFile(credentials_path)

        if gauth.credentials is None:
            gauth.GetFlow()
            gauth.flow.params.update({'approval_prompt': 'force'})
            gauth.LocalWebserverAuth()
        elif gauth.access_token_expired:
            gauth.GetFlow()
            gauth.flow.params.update({'approval_prompt': 'force'})
            gauth.LocalWebserverAuth()
        else:
            gauth.Authorize()

        gauth.SaveCredentialsFile(credentials_path)
        self.gdrive = GoogleDrive(gauth)

    def ls(self, folder_id):
        try:
            file_list = self.gdrive.ListFile({'q': f"'{folder_id}' in parents and trashed=false"}).GetList()
            result = []
            for file in file_list:
                result.append((file['title'], file['id']))
        except:
            result = None
        return result

    def get_folder_id_by_path(self, path):
        folder_names = path.strip("/").split("/")  # Split the path into folder names
        parent_id = "root"  # Start from the root directory

        for folder_name in folder_names:
            query = f"title = '{folder_name}' and mimeType = 'application/vnd.google-apps.folder' and trashed = false and '{parent_id}' in parents"
            folder_list = self.gdrive.ListFile({'q': query}).GetList()

            if not folder_list:
                return None  # Folder not found

            parent_id = folder_list[0]['id']  # Move to the next folder in the path

        return parent_id  # Return the final folder ID

    def get_shared_folder_id(self, folder_name):
        query = "title = '" + folder_name + "' and mimeType = 'application/vnd.google-apps.folder' and trashed = false"
        shared_folders = self.gdrive.ListFile(
            {'q': query, 'spaces': 'drive', 'corpora': 'allDrives', 'supportsAllDrives': True}).GetList()
        if shared_folders:
            folder_id = shared_folders[0]['id']
            return folder_id
        else:
            return None

    def upload_file(self, filename, folder_id):
        file = self.gdrive.CreateFile({'title': os.path.basename(filename), 'parents': [{'id': folder_id}]})
        file.SetContentFile(filename)
        file.Upload()
        return file['id']


class Sheets:

    def __init__(self, **kwargs):
        self.base_folder = kwargs.get("base_folder", "../..")
        self.config_dir = kwargs.get("config_dir", "../..")
        self.gc = None
        self.wb = None
        self.woksheets = None
        self.woksheets_names = None
        if kwargs.get("authorize", True):
            self.authorize()

    def authorize(self):
        self.gc = gspread.oauth(credentials_filename=self.config_dir + os.sep + "client_secret.json",
                                authorized_user_filename=self.config_dir + os.sep + "token.json")

    def open(self, args_workbook):
        self.wb = self.gc.open(args_workbook)
        self.woksheets = self.wb.worksheets()
        self.woksheets_names = [sheet.title for sheet in self.woksheets]

    def set_base_folder(self, folder):
        self.base_folder = folder

    def upload_all(self, args_filter=None, args_yes=False):
        csv_files = [self.base_folder + os.sep + f for f in os.listdir(self.base_folder) if f.endswith(".csv")]

        if args_filter is not None:
            csv_files = [f for f in csv_files if args_filter in f]

        self._upload(csv_files, args_yes=args_yes)

    def upload(self, filename, args_yes=False):
        self._upload([self.base_folder + os.sep + f for f in filename], args_yes)

    def _upload(self, csv_files, args_yes=True, sep="\t"):

        for csv_file in csv_files:

            info = csv_file.split(":")
            if len(info) == 2:
                csv_file, corner = info
            else:
                csv_file, corner = info[0], "A1"

            print("Uploading file {} at corner: {}".format(csv_file, corner))

            title = str(os.path.basename(csv_file).replace(".csv", ""))

            row, col = a1_to_rowcol(corner)

            if title in self.woksheets_names:
                if not args_yes:
                    ok = input("Sheet {} already exists. Continue (Y/n)? ".format(title))
                    if ok.lower() != "y":
                        continue
                new_ws = self.wb.worksheet(title)
                rows = max(new_ws.row_count, row)
                cols = max(new_ws.col_count, col)
                new_ws.resize(rows, cols)
            else:
                new_ws = self.wb.add_worksheet(title, row, col)

            if len(csv_file.split(".")) == 1:
                csv_file += ".csv"

            with open(csv_file, "r", encoding='utf-8') as f:
                data = f.readlines()
                data = [line.strip().split(sep) for line in data]

                for row in data:
                    for i in range(len(row)):
                        row[i] = get_narrowest_type(row[i])

            # print("Uploading sheet {}".format(title))
            new_ws.update(data, corner, value_input_option=ValueInputOption.user_entered)

    def download(self, args_sheet, args_yes=False):
        self._download(args_sheet, self.base_folder, args_yes)

    def download_all(self, args_filter=None, args_yes=False):
        # Download files
        sheets_to_download = self.woksheets_names
        if args_filter is not None:
            sheets_to_download = [sheet for sheet in sheets_to_download if args_filter in sheet]
        self._download(sheets_to_download, self.base_folder, args_yes)

    def _download(self, sheets_to_download, args_folder, args_yes=False):
        for sheet in sheets_to_download:
            ws = self.wb.worksheet(sheet)
            filename = args_folder + os.sep + str(sheet + ".csv")

            if os.path.exists(filename) and not args_yes:
                ok = input("File {} already exists. Overwrite (Y/n)? ".format(filename))
                if ok.lower() != "y":
                    continue
            with open(filename, "w", encoding='utf-8') as f:
                print("Downloading sheet {}".format(sheet))
                data = ws.get_all_values()
                for line in data:
                    f.write(",".join(line) + "\n")

    def diff(self, args_diff):
        self._diff(args_diff, self.base_folder)

    def _diff(self, args_diff, args_folder, sep="\t"):
        if args_diff:
            ws = self.wb.worksheet(args_diff)
            ws_data = ws.get_all_values()

            filename = args_folder + os.sep + str(args_diff + ".csv")
            with open(filename, "r", encoding='utf-8') as f:
                data = f.readlines()
                data = [line.strip().split(sep) for line in data]

            data_len = len(data)
            ws_len = len(ws_data)

            if data_len != ws_len:
                print("Data have different number of rows")
                print("Local:  ", data_len)
                print("Remote: ", ws_len)

            min_length = min(data_len, ws_len)

            for i in range(min_length):
                if data[i] != ws_data[i]:
                    print("Row {}:".format(i))
                    print("Local:  ", str(data[i]).replace("'", ""))
                    print("Remote: ", str(ws_data[i]).replace("'", ""))
                    print("")



import argparse
import os
import re
import sys

import yaml

from qrgrader.common import check_workspace
from qrgrader.utils import makedir

# sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from qrgrader.gdrive import GDrive, Sheets

client_secrets_json = '''{
}'''


def main():
    parser = argparse.ArgumentParser(description='Upload and download sheets from Google Sheets')
    parser.add_argument('-d', '--download', help='Download sheet', action="append", default=[])
    parser.add_argument('-D', '--download-all', help='Download sheet', action="store_true")
    parser.add_argument('-f', '--file', help='Config file')
    parser.add_argument('-F', '--filter', help='Filter names')
    parser.add_argument('-g', '--upload-pdf', help='Specify folder id for uploading', action="store_true")
    parser.add_argument('-i', '--folder-id', help='Specify folder id for uploading', default=None)
    parser.add_argument('-l', '--list', help='Create a PDF sheet', action="store_true")
    parser.add_argument('-p', '--folder-path', help='Specify folder path for uploading', default=None)
    parser.add_argument('-P', '--shared-folder-name', help='Specify folder name for uploading or listing', default=None)
    parser.add_argument('-u', '--upload', help='Upload sheet', action="append", default=[])
    parser.add_argument('-U', '--upload-all', help='Upload all sheets', action="store_true")
    parser.add_argument('-w', '--workbook', help='Workbook name')
    parser.add_argument('-x', '--diff', help='Difference')
    parser.add_argument('-y', '--yeah', help="Answer 'yes' to all questions", action="store_true")

    args = vars(parser.parse_args())

    workspace_dir = os.path.basename(os.getcwd())

    if not check_workspace():
        print("ERROR: qrsheets must be run from a workspace directory")
        sys.exit(1)

    date = workspace_dir.split("-")[1]

    # Create json file with client secrets
    makedir("config")
    if not os.path.exists("config" + os.sep + "client_secret.json"):
        with open("config" + os.sep + "client_secret.json", "w", encoding='utf-8') as f:
            f.write(client_secrets_json)

    # Load config file if provided
    if args["file"] is not None:
        with open(args["file"], "r", encoding='utf-8') as f:
            file_args = yaml.safe_load(f)
            args.update(file_args)

    if args["file"] is None and args["workbook"] is None and args["upload_pdf"] is False and args["list"] is False:
        print("You must specify at least one of the following options: --file, --workbook, --upload-pdf, --list")
        return

    if args.get("upload_pdf") or args.get("list"):

        gd = GDrive(config_dir="config")

        if args["folder_id"] is not None:
            folder_id = args["folder_id"]
        elif args["folder_path"] is not None:
            folder_id = gd.get_folder_id_by_path(args["folder_path"])
        elif args["shared_folder_name"] is not None:
            folder_id = gd.get_shared_folder_id(args["shared_folder_name"])
        else:
            print("You must specify a folder-id, folder-path or shared-folder-name")
            sys.exit(1)

        if folder_id is None:
            print("Folder does not exist, exiting.")
            sys.exit(1)

        if args.get("upload_pdf"):
            for file in os.listdir("results" + os.sep + "pdf"):
                if file.endswith(".pdf"):
                    print("Uploading file", file)
                    gd.upload_file("results" + os.sep + "pdf" + os.sep + file, folder_id)
            print("All files uploaded to folder", folder_id)
            args["list"] = folder_id

        if args.get("list", None):
            files = gd.ls(folder_id)

            with open("results" + os.sep + "xls" + os.sep + date + "_pdf.csv", "w", encoding='utf-8') as f:
                for file in files:
                    f.write(file[0].replace(".pdf", "") + "\t" + file[1] + "\n")

            print("Written to results/xls/{}_pdf.csv ({} rows)".format(date, len(files)))

    args_workbook = args.get("workbook", None)
    if args_workbook is not None:

        args_upload = args.get("upload", [])
        args_filter = args.get("filter", None)
        args_download = args.get("download", [])
        args_download_all = args.get("download_all", False)
        args_yes = args.get("yeah")
        args_upload_all = args.get("upload_all", False)
        args_diff = args.get("diff", None)

        # args_filter = date if args_filter is None else args_filter
        args_filter = None if args_filter == "." else args_filter

        if len(args_upload) > 0 and args_upload_all:
            print("You cannot use both --upload and --upload-all")
            return

        if len(args_download) > 0 and args_download_all:
            print("You cannot use both --download and --download-all")
            return

        sh = Sheets(base_folder="results/xls", config_dir="config")
        sh.open(args_workbook)

        if args_upload_all:
            sh.upload_all(args_filter, args_yes)

        elif len(args_upload) > 0:
            sh.upload(args_upload, args_yes)

        if args_download_all:
            sh.download_all(args_filter, args_yes)
        elif len(args_download) > 0:
            sh.download(args_download, args_yes)

        if args_diff:
            sh.diff(args_diff)

    print("All Done :)")


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

setup(
    name='qrgrader',
    version='0.0.1',
    packages=find_packages(where='src'),  # Specify src directory
    package_dir={'': 'src'},  # Tell setuptools that packages are under src
    install_requires=[
        'pyqt5',
        'pymupdf >= 1.18.17',
        'easyconfig2',
        'zxing-cpp',
        'gspread',
        'pydrive2',
        'opencv-python-headless',
        'pandas',
        'swikv4-minimal'
    ],
    author='Danilo Tardioli',
    author_email='dantard@unizar.es',
    description='A framwork for automatic grading of exams using QR codes',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/dantard/qrgrader',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
    entry_points={
        'console_scripts': [
            'qrscanner=qrgrader.qrscanner:main',
            'qrgrader=qrgrader.qrgui:main',
            'qrsheets=qrgrader.qrsheets:main',
            'qrgenerator=qrgrader.qrgenerator:main',
            'qrworkspace=qrgrader.qrworkspace:main',
        ],
    },
    package_data={
        "qrgrader": ["latex/*"],
    },
    include_package_data=True,
)



import shutil
import subprocess
import sys
import threading
from multiprocessing import Process

from pymupdf import pymupdf


class Done:
    def __init__(self):
        self.done = 0
        self.mutex = threading.Lock()

    def inc_done(self):
        self.mutex.acquire()
        self.done = self.done + 1
        self.mutex.release()

    def get_done(self):
        return self.done


class Generator(Process):
    def __init__(self, semaphore, filename, uniqueid, **kwargs):
        super().__init__()
        self.uniqueid = uniqueid
        self.dir_generated = kwargs.get("dir_generated", "../..")
        self.source_filename = filename
        self.dir_temp_generator = kwargs.get("dir_temp_generator", "../..")
        self.desired_pages = kwargs.get("desired_pages", -1)
        self.verbose = kwargs.get("verbose", False)
        self.dir_source = kwargs.get("dir_source", "../..")
        self.semaphore = semaphore

    def run(self):
        process = subprocess.Popen(['xelatex',
                                    '-interaction', 'nonstopmode',
                                    '-halt-on-error',
                                    '-output-directory', '{}'.format(self.dir_temp_generator),
                                    '-jobname', '{}'.format(self.uniqueid),
                                    '\\newcommand{{\\uniqueid}}{{{:s}}}\\input{{{:s}}}"'.format(self.uniqueid,
                                                                                                self.source_filename.replace("\\", "/"))],
                                   stdout=subprocess.PIPE,
                                   universal_newlines=True,
                                   cwd=self.dir_source)

        while True:
            output = process.stdout.readline()
            self.verbose and print(output.strip())
            if "MATCODE" in output.strip():
                print(output)

            # Do something else
            return_code = process.poll()
            if return_code is not None:
                if return_code == 0:
                    this_one_done = True
                    if self.desired_pages != -1:
                        the_pdf = pymupdf.open(self.dir_temp_generator + str(self.uniqueid) + ".pdf")
                        if the_pdf.page_count != self.desired_pages:
                            this_one_done = False
                            print("Discarding exam {} ({} pages)".format(self.uniqueid, the_pdf.page_count))

                    if this_one_done:
                        shutil.move(self.dir_temp_generator + str(self.uniqueid) + ".pdf", self.dir_generated + str(self.uniqueid) + ".pdf")

                else:
                    print(" * ERROR: Exam {} generation has finished with return code: {}".format(self.uniqueid, return_code))
                # Process has finished, read rest of the output
                for output in process.stdout.readlines():
                    (self.verbose or return_code != 0) and print(output.strip())

                break
        self.semaphore.release()
